import shutil
from pathlib import Path
from time import sleep

import lamindb as ln
import pytest
from lamindb.core.loaders import load_h5ad
from lamindb_setup._set_managed_storage import set_managed_storage


# https://stackoverflow.com/questions/22627659/run-code-before-and-after-each-test-in-py-test
# switch to cloud storage and back
@pytest.fixture
def switch_storage():
    cloud_storage = "s3://lamindb-ci/lamindb-unit-tests-cloud"
    set_managed_storage(cloud_storage)
    yield cloud_storage
    set_managed_storage("./default_storage_unit_storage")


def test_local_cache():
    # check that we have local storage
    local_storage = Path("./default_storage_unit_storage").resolve().as_posix()
    assert ln.setup.settings.storage.root_as_str == local_storage

    test_file = ln.examples.datasets.anndata_file_pbmc68k_test()
    adata = load_h5ad(test_file)

    artifact = ln.Artifact.from_anndata(adata, key="test_cache.h5ad")
    temp_path = artifact._local_filepath.resolve()
    assert temp_path.exists()
    assert ln.setup.settings.cache_dir in temp_path.parents

    artifact.save()
    assert artifact.path.exists()
    assert not temp_path.exists()

    artifact.delete(permanent=True)

    # check directories
    adata_zarr_pth = Path("test_adata.zarr")
    adata.write_zarr(adata_zarr_pth)
    assert adata_zarr_pth.exists()

    artifact = ln.Artifact(adata_zarr_pth, key="test_cache.zarr").save()
    assert adata_zarr_pth.exists()
    assert artifact.path.exists()
    assert artifact.path.name != artifact.key

    shutil.rmtree(adata_zarr_pth)
    artifact.delete(permanent=True)

    # check directories in cache
    cache_dir = ln.setup.settings.cache_dir
    adata_zarr_pth = cache_dir / "test_adata.zarr"
    adata.write_zarr(adata_zarr_pth)

    artifact = ln.Artifact(adata_zarr_pth, key="test_cache.zarr")
    assert adata_zarr_pth.exists()
    artifact.save()

    assert not adata_zarr_pth.exists()
    assert artifact.path.exists()
    assert artifact.path.name != artifact.key

    artifact.delete(permanent=True)


def test_cloud_cache(switch_storage):
    # check that we have cloud storage
    assert ln.setup.settings.storage.root_as_str == switch_storage

    cache_dir = ln.setup.settings.cache_dir
    assert cache_dir is not None

    test_file = ln.examples.datasets.anndata_file_pbmc68k_test()

    # test cache for saving an in-memory object
    adata = load_h5ad(test_file)

    artifact = ln.Artifact.from_anndata(adata, key="test_cache.h5ad")
    temp_path = artifact._local_filepath.resolve()
    assert cache_dir in temp_path.parents
    artifact.save()
    assert not temp_path.exists()
    cloud_path = artifact.path
    cache_path = artifact._cache_path
    assert cache_path.exists()
    assert (
        cache_path == cache_dir / "lamindb-ci/lamindb-unit-tests-cloud/test_cache.h5ad"
    )
    assert cloud_path.modified.timestamp() < cache_path.stat().st_mtime

    artifact.delete(permanent=True)

    # test cache for saving an on-disk object
    artifact = ln.Artifact.from_anndata(test_file, key="test_cache.h5ad")
    artifact.save()
    cloud_path = artifact.path
    cache_path = artifact._cache_path
    assert cache_path.exists()
    assert (
        cache_path == cache_dir / "lamindb-ci/lamindb-unit-tests-cloud/test_cache.h5ad"
    )
    assert test_file.stat().st_mtime < cache_path.stat().st_mtime
    assert cloud_path.modified.timestamp() < cache_path.stat().st_mtime

    artifact.delete(permanent=True)

    # test cache for a directory on-disk object outside the cache dir
    adata_zarr_pth = Path("test_adata.zarr")
    adata.write_zarr(adata_zarr_pth)
    artifact = ln.Artifact(adata_zarr_pth, key="test_cache.zarr")
    artifact.save()
    assert adata_zarr_pth.is_dir()
    cache_path = artifact._cache_path
    assert cache_path.is_dir()
    assert (
        cache_path == cache_dir / "lamindb-ci/lamindb-unit-tests-cloud/test_cache.zarr"
    )

    shutil.rmtree(adata_zarr_pth)
    artifact.delete(permanent=True)

    # inside the cache dir
    adata_zarr_pth = cache_dir / "test_adata.zarr"
    adata.write_zarr(adata_zarr_pth)
    artifact = ln.Artifact(adata_zarr_pth, key="test_cache.zarr")
    assert adata_zarr_pth.exists()
    artifact.save()
    assert not adata_zarr_pth.exists()
    cache_path = artifact._cache_path
    assert cache_path.is_dir()
    assert (
        cache_path == cache_dir / "lamindb-ci/lamindb-unit-tests-cloud/test_cache.zarr"
    )

    artifact.delete(permanent=True)


def test_cloud_cache_versions(switch_storage):
    adata = load_h5ad(ln.examples.datasets.anndata_file_pbmc68k_test())

    cache_dir = ln.setup.settings.cache_dir
    assert cache_dir is not None

    artifact = ln.Artifact.from_anndata(adata, key="test_cache.h5ad")
    assert ln.settings.cache_dir in artifact._local_filepath.parents
    artifact.save()
    cache_path_v1 = artifact.cache()
    assert cache_path_v1.exists()
    assert (
        cache_path_v1
        == cache_dir / "lamindb-ci/lamindb-unit-tests-cloud/test_cache.h5ad"
    )
    cache_path_v1.unlink()
    artifact.cache(print_progress=False)
    assert cache_path_v1.exists()
    assert (
        cache_path_v1
        == cache_dir / "lamindb-ci/lamindb-unit-tests-cloud/test_cache.h5ad"
    )
    timestamp_v1 = cache_path_v1.stat().st_mtime
    # hope it is enough to avoid random timestamp problems further
    sleep(1)
    # new version
    adata.obs["test_cache"] = "test"
    artifact_v2 = ln.Artifact.from_anndata(
        adata, key="test_cache.h5ad", revises=artifact
    )
    assert ln.settings.cache_dir in artifact_v2._local_filepath.parents
    artifact_v2.save()
    assert artifact_v2.is_latest
    assert not artifact.is_latest
    cache_path_v2 = artifact_v2.cache()
    assert cache_path_v2.exists()
    assert (
        cache_path_v2
        == cache_dir / "lamindb-ci/lamindb-unit-tests-cloud/test_cache.h5ad"
    )
    assert cache_path_v2.stat().st_mtime > timestamp_v1
    cache_path_v2.unlink()
    artifact_v2.cache(mute=True)
    assert cache_path_v2.exists()
    assert (
        cache_path_v2
        == cache_dir / "lamindb-ci/lamindb-unit-tests-cloud/test_cache.h5ad"
    )
    assert "test_cache" in load_h5ad(cache_path_v2).obs.columns
    cache_mtime = cache_path_v2.stat().st_mtime
    assert cache_mtime == artifact_v2.path.modified.timestamp()
    assert cache_mtime > timestamp_v1
    # old version cache ignores key
    cache_path_v1 = artifact.cache()
    assert cache_path_v1.exists()
    assert cache_path_v1.name == f"{artifact.uid}.h5ad"

    artifact_v2.versions.delete(permanent=True)


def test_corrupted_cache_local():
    filepath = ln.examples.datasets.anndata_file_pbmc68k_test()
    artifact = ln.Artifact.from_anndata(filepath, key="test_corrupt_cache_local.h5ad")
    artifact.save()
    # corrupt cache
    with open(artifact._cache_path, "r+b") as f:
        f.write(b"corruption")
    # just raises an exception, nothing to re-sync on local
    with pytest.raises(OSError):
        artifact.load()
    with pytest.raises(OSError):
        artifact.open()

    artifact.delete(permanent=True)


def test_corrupted_cache_cloud(switch_storage):
    # check that we have cloud storage
    assert ln.setup.settings.storage.root_as_str == switch_storage

    filepath = ln.examples.datasets.anndata_file_pbmc68k_test()
    artifact = ln.Artifact.from_anndata(filepath, key="test_corrupt_cache_cloud.h5ad")
    artifact.save()
    # corrupt cache
    # sleep not to reset cache mtime to a smaller value
    # it is increased artificially on cache copying in save
    # so due to lower granularity of cloud mtimes and fast code execution
    # after the change cache mtime can become smaller than cloud mtime
    sleep(1)
    with open(artifact._cache_path, "r+b") as f:
        f.write(b"corruption")
    assert artifact._cache_path.stat().st_mtime > artifact.path.stat().st_mtime
    # check that it is indeed corrupted
    with pytest.raises(OSError):
        load_h5ad(artifact.cache())
    # should load successfully
    artifact.load()
    # check open also
    assert artifact._cache_path.exists()
    with open(artifact._cache_path, "r+b") as f:
        f.write(b"corruption")
    # should open successfully
    with artifact.open():
        pass
    # corrupted cache has been deleted
    assert not artifact._cache_path.exists()

    artifact.delete(permanent=True)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, overload

import anndata as ad
import pandas as pd
from django.db import models
from django.db.models import CASCADE, PROTECT, Q
from lamin_utils import logger
from lamindb_setup.core.hashing import HASH_LENGTH, hash_set

from lamindb.base.fields import (
    CharField,
    ForeignKey,
    OneToOneField,
    TextField,
)

from ..base.ids import base62_20
from ..core._mapped_collection import MappedCollection
from ..core.storage._backed_access import _open_dataframe
from ..errors import FieldValidationError
from ..models._is_versioned import process_revises
from ._is_versioned import IsVersioned
from .artifact import (
    Artifact,
    describe_artifact_collection,
    get_run,
    populate_subsequent_run,
    save_schema_links,
    track_run_input,
)
from .has_parents import view_lineage
from .run import Run, TracksRun, TracksUpdates
from .sqlrecord import (
    BaseSQLRecord,
    IsLink,
    SQLRecord,
    _get_record_kwargs,
    init_self_from_db,
    update_attributes,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from polars import LazyFrame as PolarsLazyFrame
    from pyarrow.dataset import Dataset as PyArrowDataset

    from ..core.storage import UPath
    from .block import CollectionBlock
    from .project import Project, Reference
    from .query_set import QuerySet
    from .transform import Transform
    from .ulabel import ULabel


# below is a draft for the future, see also the tests in test_collection.py
#
# class CollectionFeatureManager:
#     """Query features of artifact in collection."""

#     def __init__(self, collection: Collection):
#         self._collection = collection

#     def _get_staged_feature_sets_union(self) -> dict[str, Schema]:
#         links_schema_artifact = Artifact.feature_sets.through.objects.filter(
#             artifact_id__in=self._collection.artifacts.values_list("id", flat=True)
#         )
#         feature_sets_by_slots = defaultdict(list)
#         for link in links_schema_artifact:
#             feature_sets_by_slots[link.slot].append(link.schema_id)
#         feature_sets_union = {}
#         for slot, schema_ids_slot in feature_sets_by_slots.items():
#             schema_1 = Schema.get(id=schema_ids_slot[0])
#             related_name = schema_1._get_related_name()
#             features_registry = getattr(Schema, related_name).field.model
#             # this way of writing the __in statement turned out to be the fastest
#             # evaluated on a link table with 16M entries connecting 500 feature sets with
#             # 60k genes
#             feature_ids = (
#                 features_registry.schemas.through.objects.filter(
#                     schema_id__in=schema_ids_slot
#                 )
#                 .values(f"{features_registry.__name__.lower()}_id")
#                 .distinct()
#             )
#             features = features_registry.filter(id__in=feature_ids)
#             feature_sets_union[slot] = Schema(features, dtype=schema_1.dtype)
#         return feature_sets_union


def _load_concat_artifacts(
    artifacts: list[Artifact], join: Literal["inner", "outer"] = "outer", **kwargs
) -> pd.DataFrame | ad.AnnData:
    suffixes = {artifact.suffix for artifact in artifacts}
    # Why is that? - Sergei
    if len(suffixes) != 1:
        raise ValueError(
            "Can only load collections where all artifacts have the same suffix"
        )

    # because we're tracking data flow on the collection-level, here, we don't
    # want to track it on the artifact-level
    first_object = artifacts[0].load(is_run_input=False)
    is_dataframe = isinstance(first_object, pd.DataFrame)
    is_anndata = isinstance(first_object, ad.AnnData)
    if not is_dataframe and not is_anndata:
        raise ValueError(f"Unable to concatenate {suffixes.pop()} objects.")

    objects = [first_object]
    artifact_uids = [artifacts[0].uid]
    for artifact in artifacts[1:]:
        objects.append(artifact.load(is_run_input=False))
        artifact_uids.append(artifact.uid)

    if is_dataframe:
        concat_object = pd.concat(objects, join=join, **kwargs)
    elif is_anndata:
        label = kwargs.pop("label", "artifact_uid")
        keys = kwargs.pop("keys", artifact_uids)
        concat_object = ad.concat(objects, join=join, label=label, keys=keys, **kwargs)
    return concat_object


class Collection(SQLRecord, IsVersioned, TracksRun, TracksUpdates):
    """Versioned collections of artifacts.

    Args:
        artifacts: `Artifact | list[Artifact]` One or several artifacts.
        key: `str` A file-path like key, analogous to the `key` parameter of `Artifact` and `Transform`.
        description: `str | None = None` A description.
        revises: `Collection | None = None` An old version of the collection.
        run: `Run | None = None` The run that creates the collection.
        meta: `Artifact | None = None` An artifact that defines metadata for the collection.
        reference: `str | None = None` A simple reference, e.g. an external ID or a URL.
        reference_type: `str | None = None` A way to indicate to indicate the type of the simple reference `"url"`.

    See Also:
        :class:`~lamindb.Artifact`

    Examples:

        Create a collection from a list of :class:`~lamindb.Artifact` objects:

        >>> collection = ln.Collection([artifact1, artifact2], key="my_project/my_collection")

        Create a collection that groups a data & a metadata artifact (e.g., here :doc:`docs:rxrx`):

        >>> collection = ln.Collection(data_artifact, key="my_project/my_collection", meta=metadata_artifact)

    """

    class Meta(SQLRecord.Meta, IsVersioned.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "lamindb"
        constraints = [
            models.UniqueConstraint(
                fields=["key", "hash"],
                name="unique_collection_key_hash_not_null",
            )
        ]

    _len_full_uid: int = 20
    _len_stem_uid: int = 16
    _name_field: str = "key"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(
        editable=False,
        unique=True,
        db_index=True,
        max_length=_len_full_uid,
        default=base62_20,
    )
    """Universal id, valid across DB instances."""
    key: str = CharField(db_index=True)
    """Name or path-like key."""
    # below is the only case in which we use a TextField
    # for description; we do so because users had descriptions exceeding 255 chars
    # in their instances
    description: str | None = TextField(null=True)
    """A description or title."""
    hash: str | None = CharField(
        max_length=HASH_LENGTH,
        db_index=True,
        null=True,
    )
    """Hash of collection content."""
    reference: str | None = CharField(max_length=255, db_index=True, null=True)
    """A reference like URL or external ID."""
    # also for reference_type here, we allow an extra long max_length
    reference_type: str | None = CharField(max_length=25, db_index=True, null=True)
    """Type of reference, e.g., cellxgene Census collection_id."""
    ulabels: ULabel = models.ManyToManyField(
        "ULabel", through="CollectionULabel", related_name="collections"
    )
    """ULabels annotating the collection (see :class:`~lamindb.Feature`)."""
    run: Run | None = ForeignKey(
        Run, PROTECT, related_name="output_collections", null=True, default=None
    )
    """:class:`~lamindb.Run` that created the `collection`."""
    input_of_runs: Run = models.ManyToManyField(Run, related_name="input_collections")
    """Runs that use this collection as an input."""
    _subsequent_runs: Run = models.ManyToManyField(
        "Run",
        related_name="_recreated_collections",
        db_table="lamindb_collection__previous_runs",  # legacy name, change in lamindb v2
    )
    """Runs that re-created the record after initial creation."""
    artifacts: Artifact = models.ManyToManyField(
        "Artifact", related_name="collections", through="CollectionArtifact"
    )
    """Artifacts in collection."""
    meta_artifact: Artifact | None = OneToOneField(
        "Artifact",
        PROTECT,
        null=True,
        unique=True,
        related_name="_meta_of_collection",
    )
    """An artifact that stores metadata that indexes a collection.

    It has a 1:1 correspondence with an artifact. If needed, you can access the
    collection from the artifact via a private field:
    `artifact._meta_of_collection`.
    """
    _actions: Artifact = models.ManyToManyField(Artifact, related_name="+")
    """Actions to attach for the UI."""
    projects: Project
    """Linked projects."""
    references: Reference
    """Linked references."""
    blocks: CollectionBlock
    """Blocks that annotate this collection."""

    @overload
    def __init__(
        self,
        artifacts: Artifact | list[Artifact],
        key: str,
        description: str | None = None,
        meta: Any | None = None,
        reference: str | None = None,
        reference_type: str | None = None,
        run: Run | None = None,
        revises: Collection | None = None,
        skip_hash_lookup: bool = False,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        if len(args) == len(self._meta.concrete_fields):
            super().__init__(*args, **kwargs)
            return None
        # now we proceed with the user-facing constructor
        if len(args) > 1:
            raise ValueError("Only one non-keyword arg allowed: artifacts")
        artifacts: Artifact | list[Artifact] = (
            kwargs.pop("artifacts") if len(args) == 0 else args[0]
        )
        meta_artifact: Artifact | None = kwargs.pop("meta_artifact", None)
        tmp_key: str | None = kwargs.pop("key", None)
        description: str | None = kwargs.pop("description", None)
        reference: str | None = kwargs.pop("reference", None)
        reference_type: str | None = kwargs.pop("reference_type", None)
        run: Run | None = kwargs.pop("run", None)
        revises: Collection | None = kwargs.pop("revises", None)
        version: str | None = kwargs.pop("version", None)
        skip_hash_lookup: bool = kwargs.pop("skip_hash_lookup", False)
        branch = kwargs.pop("branch", None)
        branch_id = kwargs.pop("branch_id", 1)
        space = kwargs.pop("space", None)
        space_id = kwargs.pop("space_id", 1)
        key: str
        if "name" in kwargs:
            key = kwargs.pop("name")
            warnings.warn(
                f"argument `name` will be removed, please pass {key} to `key` instead",
                FutureWarning,
                stacklevel=2,
            )
        else:
            key = tmp_key
        if not len(kwargs) == 0:
            valid_keywords = ", ".join(
                [val[0] for val in _get_record_kwargs(Collection)]
            )
            raise FieldValidationError(
                f"Only {valid_keywords} can be passed, you passed: {kwargs}"
            )
        if revises is None:
            revises = (
                Collection.filter(key=key, is_latest=True)
                .order_by("-created_at")
                .first()
            )
        provisional_uid, version, key, description, revises = process_revises(
            revises, version, key, description, Collection
        )
        run = get_run(run)
        if isinstance(artifacts, Artifact):
            artifacts = [artifacts]
        else:
            if not hasattr(artifacts, "__getitem__"):
                raise ValueError("Artifact or list[Artifact] is allowed.")
            assert isinstance(artifacts[0], Artifact)  # type: ignore  # noqa: S101
        hash = from_artifacts(artifacts)  # type: ignore
        if meta_artifact is not None:
            if not isinstance(meta_artifact, Artifact):
                raise ValueError("meta_artifact has to be an Artifact")
            if isinstance(meta_artifact, Artifact):
                if meta_artifact._state.adding:
                    raise ValueError(
                        "Save meta_artifact artifact before creating collection!"
                    )
        # we ignore collections in trash containing the same hash
        if hash is not None and not skip_hash_lookup:
            existing_collection = Collection.objects.filter(
                ~Q(branch_id=-1),
                hash=hash,
            ).one_or_none()
        else:
            existing_collection = None
        if existing_collection is not None:
            logger.warning(
                f"returning collection with same hash: {existing_collection}; if you intended to query to track this collection as an input, use: ln.Collection.get()"
            )
            init_self_from_db(self, existing_collection)
            update_attributes(self, {"description": description, "key": key})
            populate_subsequent_run(self, run)
        else:
            _skip_validation = revises is not None and key == revises.key
            super().__init__(  # type: ignore
                uid=provisional_uid,
                key=key,
                description=description,
                reference=reference,
                reference_type=reference_type,
                meta_artifact=meta_artifact,
                hash=hash,
                run=run,
                version=version,
                branch=branch,
                branch_id=branch_id,
                space=space,
                space_id=space_id,
                revises=revises,
                _skip_validation=_skip_validation,
            )
        self._artifacts = artifacts
        if revises is not None and revises.uid != self.uid:
            track_run_input(revises, run=run)
        track_run_input(artifacts, run=run)

    @classmethod
    def get(
        cls,
        idlike: int | str | None = None,
        *,
        is_run_input: bool | Run = False,
        **expressions,
    ) -> Artifact:
        """Get a single collection.

        Args:
            idlike: Either a uid stub, uid or an integer id.
            is_run_input: Whether to track this collection as run input.
            expressions: Fields and values passed as Django query expressions.

        Raises:
            :exc:`docs:lamindb.errors.DoesNotExist`: In case no matching record is found.

        See Also:
            - Method in `SQLRecord` base class: :meth:`~lamindb.models.SQLRecord.get`

        Examples:

            ::

                collection = ln.Collection.get("okxPW6GIKBfRBE3B0000")
                collection = ln.Collection.get(key="scrna/collection1")
        """
        from .query_set import QuerySet

        return QuerySet(model=cls).get(idlike, is_run_input=is_run_input, **expressions)

    def append(self, artifact: Artifact, run: Run | None = None) -> Collection:
        """Append an artifact to the collection.

        This does not modify the original collection in-place, but returns a new version
        of the original collection with the appended artifact.

        Args:
            artifact: An artifact to add to the collection.
            run: The run that creates the new version of the collection.

        Examples:

            ::

                collection_v1 = ln.Collection(artifact, key="My collection").save()
                collection_v2 = collection.append(another_artifact)  # returns a new version of the collection
                collection_v2.save()  # save the new version

        """
        return Collection(  # type: ignore
            self.artifacts.all().to_list() + [artifact],
            # key is automatically derived from revises.key
            description=self.description,
            revises=self,
            run=run,
        )

    def open(
        self,
        engine: Literal["pyarrow", "polars"] = "pyarrow",
        is_run_input: bool | None = None,
        **kwargs,
    ) -> PyArrowDataset | Iterator[PolarsLazyFrame]:
        """Open a dataset for streaming.

        Works for `pyarrow` and `polars` compatible formats
        (`.parquet`, `.csv`, `.ipc` etc. files or directories with such files).

        Args:
            engine: Which module to use for lazy loading of a dataframe
                from `pyarrow` or `polars` compatible formats.
            is_run_input: Whether to track this artifact as run input.
            **kwargs: Keyword arguments for `pyarrow.dataset.dataset` or `polars.scan_*` functions.

        Notes:
            For more info, see guide: :doc:`/arrays`.
        """
        if self._state.adding:
            artifacts = self._artifacts
            logger.warning("the collection isn't saved, consider calling `.save()`")
        else:
            artifacts = self.ordered_artifacts.all()
        paths = [artifact.path for artifact in artifacts]

        dataframe = _open_dataframe(paths, engine=engine, **kwargs)
        # track only if successful
        track_run_input(self, is_run_input)
        return dataframe

    def mapped(
        self,
        layers_keys: str | list[str] | None = None,
        obs_keys: str | list[str] | None = None,
        obsm_keys: str | list[str] | None = None,
        obs_filter: dict[str, str | list[str]] | None = None,
        join: Literal["inner", "outer"] | None = "inner",
        encode_labels: bool | list[str] = True,
        unknown_label: str | dict[str, str] | None = None,
        cache_categories: bool = True,
        parallel: bool = False,
        dtype: str | None = None,
        stream: bool = False,
        is_run_input: bool | None = None,
    ) -> MappedCollection:
        """Return a map-style dataset.

        Returns a `pytorch map-style dataset
        <https://pytorch.org/docs/stable/data.html#map-style-datasets>`__ by
        virtually concatenating `AnnData` arrays.

        By default (`stream=False`) `AnnData` arrays are moved into a local
        cache first.

        `__getitem__` of the `MappedCollection` object takes a single integer index
        and returns a dictionary with the observation data sample for this index from
        the `AnnData` objects in the collection. The dictionary has keys for `layers_keys`
        (`.X` is in `"X"`), `obs_keys`, `obsm_keys` (under `f"obsm_{key}"`) and also `"_store_idx"`
        for the index of the `AnnData` object containing this observation sample.

        .. note::

            For a guide, see :doc:`docs:scrna-mappedcollection`.

            This method currently only works for collections or query sets of `AnnData` artifacts.

        Args:
            layers_keys: Keys from the ``.layers`` slot. ``layers_keys=None`` or ``"X"`` in the list
                retrieves ``.X``.
            obs_keys: Keys from the ``.obs`` slots.
            obsm_keys: Keys from the ``.obsm`` slots.
            obs_filter: Select only observations with these values for the given obs columns.
                Should be a dictionary with obs column names as keys
                and filtering values (a string or a list of strings) as values.
            join: `"inner"` or `"outer"` virtual joins. If ``None`` is passed,
                does not join.
            encode_labels: Encode labels into integers.
                Can be a list with elements from ``obs_keys``.
            unknown_label: Encode this label to -1.
                Can be a dictionary with keys from ``obs_keys`` if ``encode_labels=True``
                or from ``encode_labels`` if it is a list.
            cache_categories: Enable caching categories of ``obs_keys`` for faster access.
            parallel: Enable sampling with multiple processes.
            dtype: Convert numpy arrays from ``.X``, ``.layers`` and ``.obsm``
            stream: Whether to stream data from the array backend.
            is_run_input: Whether to track this collection as run input.

        Examples:
            >>> import lamindb as ln
            >>> from torch.utils.data import DataLoader
            >>> ds = ln.Collection.get(description="my collection")
            >>> mapped = collection.mapped(obs_keys=["cell_type", "batch"])
            >>> dl = DataLoader(mapped, batch_size=128, shuffle=True)
            >>> # also works for query sets of artifacts, '...' represents some filtering condition
            >>> # additional filtering on artifacts of the collection
            >>> mapped = collection.artifacts.all().filter(...).order_by("-created_at").mapped()
            >>> # or directly from a query set of artifacts
            >>> mapped = ln.Artifact.filter(..., otype="AnnData").order_by("-created_at").mapped()
        """
        path_list = []
        if self._state.adding:
            artifacts = self._artifacts
            logger.warning("the collection isn't saved, consider calling `.save()`")
        else:
            artifacts = self.ordered_artifacts.all()
        for artifact in artifacts:
            if ".h5ad" not in artifact.suffix and ".zarr" not in artifact.suffix:
                logger.warning(f"ignoring artifact with suffix {artifact.suffix}")
                continue
            elif not stream:
                path_list.append(artifact.cache())
            else:
                path_list.append(artifact.path)
        ds = MappedCollection(
            path_list,
            layers_keys,
            obs_keys,
            obsm_keys,
            obs_filter,
            join,
            encode_labels,
            unknown_label,
            cache_categories,
            parallel,
            dtype,
        )
        # track only if successful
        track_run_input(self, is_run_input)
        return ds

    def cache(self, is_run_input: bool | None = None) -> list[UPath]:
        """Download cloud artifacts in collection to local cache.

        Follows syncing logic: only downloads outdated artifacts.

        Returns ordered paths to locally cached on-disk artifacts via `.ordered_artifacts.all()`:

        Args:
            is_run_input: Whether to track this collection as run input.
        """
        path_list = []
        for artifact in self.ordered_artifacts.all():
            # do not want to track data lineage on the artifact level
            path_list.append(artifact.cache(is_run_input=False))
        track_run_input(self, is_run_input)
        return path_list

    def load(
        self,
        join: Literal["inner", "outer"] = "outer",
        is_run_input: bool | None = None,
        **kwargs,
    ) -> pd.DataFrame | ad.AnnData:
        """Cache and load to memory.

        Returns an in-memory concatenated `DataFrame` or `AnnData` object.
        """
        # cannot call track_run_input here, see comment further down
        artifacts = self.ordered_artifacts.all()
        concat_object = _load_concat_artifacts(artifacts, join, **kwargs)
        # only call it here because there might be errors during load or concat
        track_run_input(self, is_run_input)
        return concat_object

    def save(self, using: str | None = None) -> Collection:
        """Save the collection and underlying artifacts to database & storage.

        Args:
            using: The database to which you want to save.

        Examples:
            >>> collection = ln.Collection("./myfile.csv", name="myfile")
        """
        if self.meta_artifact is not None:
            self.meta_artifact.save()
        super().save()
        # we don't allow updating the collection of artifacts
        # if users want to update the set of artifacts, they
        # have to create a new collection
        if hasattr(self, "_artifacts"):
            links = [
                CollectionArtifact(collection_id=self.id, artifact_id=artifact.id)  # type: ignore
                for artifact in self._artifacts
            ]
            # the below seems to preserve the order of the list in the
            # auto-incrementing integer primary
            # merely using .artifacts.set(*...) doesn't achieve this
            # we need ignore_conflicts=True so that this won't error if links already exist
            CollectionArtifact.objects.bulk_create(links, ignore_conflicts=True)
        save_schema_links(self)
        if using is not None:
            logger.warning("using argument is ignored")
        return self

    def restore(self) -> None:
        """Restore collection record from trash.

        Examples:

            For any `Collection` object `collection`, call:

            >>> collection.restore()
        """
        self.branch_id = 1
        self.save()

    @property
    def transform(self) -> Transform | None:
        """Transform whose run created the collection."""
        return self.run.transform if self.run is not None else None

    @property
    def name(self) -> str:
        """Name of the collection.

        Splits `key` on `/` and returns the last element.
        """
        return self.key.split("/")[-1]

    @property
    def ordered_artifacts(self) -> QuerySet:
        """Ordered `QuerySet` of `.artifacts`.

        Accessing the many-to-many field `collection.artifacts` directly gives
        you non-deterministic order.

        Using the property `.ordered_artifacts` allows to iterate through a set
        that's ordered by the order of the list that created the collection.
        """
        return self.artifacts.order_by("links_collection__id")

    @property
    def data_artifact(self) -> Artifact | None:
        """Access to a single data artifact.

        If the collection has a single data & metadata artifact, this allows access via::

           collection.data_artifact  # first & only element of collection.artifacts
           collection.meta_artifact  # metadata

        """
        return self.artifacts.first()

    def describe(self) -> None:
        """Describe relations of record.

        Examples:
            >>> artifact.describe()
        """
        return describe_artifact_collection(self)


# internal function, not exposed to user
def from_artifacts(artifacts: Iterable[Artifact]) -> tuple[str, dict[str, str]]:
    # assert all artifacts are already saved
    saved = not any(artifact._state.adding for artifact in artifacts)
    if not saved:
        raise ValueError("Not all artifacts are yet saved, please save them")
    # validate consistency of hashes - we do not allow duplicate hashes
    hashes = [artifact.hash for artifact in artifacts if artifact.hash is not None]
    hashes_set = set(hashes)
    if len(hashes) != len(hashes_set):
        seen = set()
        non_unique = [x for x in hashes if x in seen or seen.add(x)]  # type: ignore
        raise ValueError(
            "Please pass artifacts with distinct hashes: these ones are non-unique"
            f" {non_unique}"
        )
    hash = hash_set(hashes_set)
    return hash


class CollectionArtifact(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    collection: Collection = ForeignKey(
        Collection, CASCADE, related_name="links_artifact"
    )
    artifact: Artifact = ForeignKey(Artifact, PROTECT, related_name="links_collection")

    class Meta:
        app_label = "lamindb"
        unique_together = ("collection", "artifact")


# mypy: ignore-errors
Collection.view_lineage = view_lineage



"""Base library.

Is available also when no instance is setup.

Modules:

.. autosummary::
   :toctree: .

   uids
   types
   fields
   dtypes

Utils:

.. autosummary::
   :toctree: .

   doc_args
   deprecated

"""

from lamindb_setup.core import deprecated, doc_args

from . import dtypes, fields, types, uids



# Generated by Django 5.2 on 2025-02-17 11:21

import django.db.models.deletion
import django.db.models.functions.datetime
from django.db import migrations, models

import lamindb.base.fields
import lamindb.base.users
import lamindb.models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0087_rename__schemas_m2m_artifact_feature_sets_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="artifactschema",
            name="slot",
            field=lamindb.base.fields.CharField(
                blank=True, default=None, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="composite",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.schema",
            ),
        ),
        migrations.CreateModel(
            name="SchemaComponent",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "slot",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255, null=True
                    ),
                ),
                (
                    "component",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_component",
                        to="lamindb.schema",
                    ),
                ),
                (
                    "composite",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_composite",
                        to="lamindb.schema",
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("composite", "component")},
            },
            bases=(lamindb.models.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="schema",
            name="components",
            field=models.ManyToManyField(
                related_name="composites",
                through="lamindb.SchemaComponent",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="artifactschema",
            unique_together={("artifact", "schema"), ("artifact", "slot")},
        ),
        migrations.AlterUniqueTogether(
            name="schemacomponent",
            unique_together={("composite", "component"), ("composite", "slot")},
        ),
        # unrelated, not on the SQL level, just switching "editable"
        migrations.AlterField(
            model_name="artifact",
            name="_hash_type",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=30,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="created_artifacts",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=22,
                null=True,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="n_files",
            field=lamindb.base.fields.BigIntegerField(
                blank=True, db_index=True, default=None, editable=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="n_observations",
            field=lamindb.base.fields.BigIntegerField(
                blank=True, db_index=True, default=None, editable=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="otype",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=64,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="output_artifacts",
                to="lamindb.run",
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="size",
            field=lamindb.base.fields.BigIntegerField(
                blank=True, db_index=True, default=None, editable=False, null=True
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="storage",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifacts",
                to="lamindb.storage",
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="suffix",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, editable=False, max_length=30
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="dtype",
            field=lamindb.base.fields.CharField(
                blank=True, default=None, editable=False, max_length=64, null=True
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=22,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="itype",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_index=True,
                default=None,
                editable=False,
                max_length=120,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="maximal_set",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, editable=False
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="minimal_set",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=True, editable=False
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="ordered_set",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False, editable=False
            ),
        ),
    ]



# Generated by Django 5.1.12 on 2025-09-28 20:30


import django.db.models.deletion
import django.db.models.functions.datetime
from django.db import migrations, models

import lamindb.base.fields
import lamindb.base.users
import lamindb.models.run
import lamindb.models.sqlrecord


def migrate_person_to_record(apps, schema_editor):
    """Migrate Person to Record."""
    Person = apps.get_model("lamindb", "Person")
    if not Person.objects.exists():
        return  # No Person records to migrate

    with schema_editor.connection.cursor() as cursor:
        cursor.execute("""
            INSERT INTO lamindb_record (
                uid, name, created_at, updated_at, created_by_id, run_id, is_type
            )
            SELECT
                CONCAT(uid, uid) as uid,
                name,
                created_at,
                updated_at,
                created_by_id,
                run_id,
                False
            FROM lamindb_person
        """)

        migrated_count = cursor.rowcount
        print(f"\nMigrated {migrated_count} Person records to Record table")

    with schema_editor.connection.cursor() as cursor:
        cursor.execute("""
            INSERT INTO lamindb_referencerecord (
                created_at, created_by_id, feature_id, record_id, reference_id, run_id
            )
            SELECT
                NOW() as created_at,
                r.created_by_id,
                NULL as feature_id,
                rec.id as record_id,
                ra.reference_id,
                rec.run_id
            FROM lamindb_reference_authors ra
            JOIN lamindb_reference r ON ra.reference_id = r.id
            JOIN lamindb_person p ON ra.person_id = p.id
            JOIN lamindb_record rec ON rec.name = p.name
        """)

        created_count = cursor.rowcount
        print(f"\nCreated {created_count} ReferenceRecord links for migrated authors")


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0121_recorduser"),
    ]

    operations = [
        migrations.CreateModel(
            name="ReferenceRecord",
            fields=[
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_referencerecord",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_reference",
                        to="lamindb.record",
                    ),
                ),
                (
                    "reference",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_record",
                        to="lamindb.reference",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.run.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("reference", "feature", "record")},
            },
            bases=(lamindb.models.sqlrecord.IsLink, models.Model),
        ),
        migrations.AddField(
            model_name="reference",
            name="records",
            field=models.ManyToManyField(
                related_name="references",
                through="lamindb.ReferenceRecord",
                to="lamindb.record",
            ),
        ),
        migrations.RunPython(migrate_person_to_record),
        migrations.RemoveField(
            model_name="project",
            name="people",
        ),
        migrations.RemoveField(
            model_name="reference",
            name="authors",
        ),
        migrations.AlterUniqueTogether(
            name="personproject",
            unique_together=None,
        ),
        migrations.AlterUniqueTogether(
            name="recordperson",
            unique_together=None,
        ),
        migrations.RenameField(
            model_name="reference",
            old_name="records",
            new_name="linked_in_records",
        ),
        migrations.AddField(
            model_name="project",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="reference",
            name="description",
            field=lamindb.base.fields.TextField(blank=True, default=None, null=True),
        ),
        migrations.DeleteModel(
            name="Person",
        ),
        migrations.DeleteModel(
            name="PersonProject",
        ),
        migrations.DeleteModel(
            name="RecordPerson",
        ),
        migrations.AlterField(
            model_name="recordreference",
            name="value",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_in_record",
                to="lamindb.reference",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="projectrecord",
            unique_together={("project", "feature", "record")},
        ),
        migrations.AddField(
            model_name="reference",
            name="records",
            field=models.ManyToManyField(
                related_name="references",
                through="lamindb.ReferenceRecord",
                to="lamindb.record",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="linked_in_records",
            field=models.ManyToManyField(
                related_name="linked_references",
                through="lamindb.RecordReference",
                to="lamindb.record",
            ),
        ),
    ]



# Generated by Django 5.1.7 on 2025-05-10 00:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0093_alter_schemacomponent_unique_together"),
    ]

    operations = [
        migrations.CreateModel(
            name="WriteLogLock",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("locked", models.BooleanField()),
            ],
        ),
        migrations.CreateModel(
            name="MigrationState",
            fields=[
                ("id", models.SmallAutoField(primary_key=True, serialize=False)),
                ("migration_state_id", models.JSONField()),
            ],
        ),
        migrations.CreateModel(
            name="TableState",
            fields=[
                ("id", models.SmallAutoField(primary_key=True, serialize=False)),
                ("table_name", models.CharField(max_length=255)),
                ("backfilled", models.BooleanField()),
            ],
        ),
        migrations.CreateModel(
            name="WriteLog",
            fields=[
                ("seqno", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        db_index=True, editable=False, max_length=18, unique=True
                    ),
                ),
                ("space_uid", models.CharField(max_length=12, null=True)),
                ("created_by_uid", models.CharField(default="00000000", max_length=8)),
                ("branch_code", models.IntegerField(default=1)),
                (
                    "run_uid",
                    models.CharField(default="0000000000000000", max_length=16),
                ),
                ("record_uid", models.JSONField(null=True)),
                ("record_data", models.JSONField(null=True)),
                ("event_type", models.PositiveSmallIntegerField()),
                ("created_at", models.DateTimeField()),
                (
                    "migration_state",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.migrationstate",
                    ),
                ),
                (
                    "table",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.tablestate",
                    ),
                ),
            ],
            options={
                "verbose_name": "Write Log",
                "verbose_name_plural": "Write Logs",
            },
        ),
    ]



import lamindb as ln

spatialdata = ln.examples.datasets.spatialdata_blobs()
sdata_schema = ln.Schema.get(name="spatialdata_blobs_schema")
curator = ln.curators.SpatialDataCurator(spatialdata, sdata_schema)
try:
    curator.validate()
except ln.errors.ValidationError:
    pass

spatialdata.tables["table"].var.drop(index="ENSG00000999999", inplace=True)

# validate again (must pass now) and save artifact
artifact = ln.Artifact.from_spatialdata(
    spatialdata, key="examples/spatialdata1.zarr", schema=sdata_schema
).save()
artifact.describe()



from __future__ import annotations

import datetime
import re
from typing import TYPE_CHECKING

from lamin_utils import logger
from rich.text import Text
from rich.tree import Tree

if TYPE_CHECKING:
    from lamindb.models import Artifact, Collection, Run


def highlight_time(iso: str):
    try:
        tz = datetime.datetime.now().astimezone().tzinfo
        res = (
            datetime.datetime.fromisoformat(iso)
            .replace(tzinfo=datetime.timezone.utc)
            .astimezone(tz)
            .strftime("%Y-%m-%d %H:%M:%S")
        )
    except ValueError:
        # raises ValueError: Invalid isoformat string: '<django.db.models.expressions.DatabaseDefault object at 0x1128ac440>'
        # but can't be caught with `isinstance(iso, DatabaseDefault)` for unkown reasons
        return Text("timestamp of unsaved record not available", style="dim")
    return Text(res)


# Define consistent column widths
NAME_WIDTH = 30
TYPE_WIDTH = 35  # types can get long, e.g. cat[Record[Treatment]]
VALUES_WIDTH = 40


def format_rich_tree(
    tree: Tree, fallback: str = "", return_str: bool = False, strip_ansi: bool = True
) -> str | None:
    from rich.console import Console

    from ..core._context import is_run_from_ipython

    # If tree has no children, return fallback
    if not tree.children:
        return fallback

    console = Console(force_terminal=True)
    printed = False

    if return_str:
        from io import StringIO

        string_io = StringIO()
        str_console = Console(file=string_io, force_terminal=True)
        str_console.print(tree)
        result = string_io.getvalue()
        if strip_ansi:
            ansi_escape = re.compile(r"\x1b(?:\[[0-9;]*[a-zA-Z]|\(B)")
            result = ansi_escape.sub("", result)
        # rstrip trailing whitespace on every line
        result = "\n".join(line.rstrip() for line in result.splitlines())
        return result

    try:
        if not is_run_from_ipython:
            from IPython import get_ipython
            from IPython.core.interactiveshell import InteractiveShell
            from IPython.display import display

            shell = get_ipython()
            if isinstance(shell, InteractiveShell):
                display(tree)
                printed = True
                return None
    except (NameError, ImportError):
        pass

    if not printed:
        # be careful to test this on a terminal
        console = Console(force_terminal=True)
        console.print(tree)

    return None


def describe_header(self: Artifact | Collection | Run) -> Tree:
    if hasattr(self, "is_latest") and not self.is_latest:
        logger.warning(
            f"This is not the latest version of the {self.__class__.__name__}."
        )
    if hasattr(self, "branch_id"):
        if self.branch_id == 0:  # type: ignore
            logger.warning("This artifact is archived.")
        elif self.branch_id == -1:  # type: ignore
            logger.warning("This artifact is in the trash.")
    # initialize tree
    suffix = self.suffix if hasattr(self, "suffix") and self.suffix else ""
    accessor = self.otype if hasattr(self, "otype") and self.otype else ""
    kind = f" · {self.kind}" if hasattr(self, "kind") and self.kind else ""
    suffix_accessor = (
        f"{suffix} · {accessor}{kind}"
        if suffix and accessor
        else suffix or accessor or ""
    )

    tree = Tree(
        Text.assemble(
            (self.__class__.__name__, "bold"), (f" {suffix_accessor}", "bold dim")
        ),
        guide_style="dim",  # dim the connecting lines
    )
    return tree


def format_bytes(bytes_value):
    """Convert bytes to human readable format."""
    if bytes_value < 1024:
        return f"{bytes_value} B"
    elif bytes_value < 1024**2:
        return f"{bytes_value / 1024:.1f} KB"
    elif bytes_value < 1024**3:
        return f"{bytes_value / (1024**2):.1f} MB"
    elif bytes_value < 1024**4:
        return f"{bytes_value / (1024**3):.1f} GB"
    else:
        return f"{bytes_value / (1024**4):.1f} TB"


def describe_artifact_general(
    self: Artifact,
    tree: Tree | None = None,
    foreign_key_data: dict[str, dict[str, int | str]] | None = None,
) -> Tree:
    if tree is None:
        tree = describe_header(self)

    # add general information (order is the same as in API docs)
    general = tree.add(Text("General", style="bold bright_cyan"))

    if self.key:
        general.add(Text.assemble(("key: ", "dim"), (f"{self.key}", "cyan3")))
    if self.description:
        general.add(
            Text.assemble(
                ("description: ", "dim"),
                f"{self.description}",
            )
        )

    # Two column items (short content)
    two_column_items = []

    two_column_items.append(Text.assemble(("uid: ", "dim"), f"{self.uid}"))
    two_column_items.append(Text.assemble(("hash: ", "dim"), f"{self.hash}"))
    two_column_items.append(
        Text.assemble(("size: ", "dim"), f"{format_bytes(self.size)}")
    )
    transform_key = (
        foreign_key_data["run"]["transform_key"]
        if foreign_key_data
        else self.transform.key
        if self.transform
        else None
    )
    two_column_items.append(
        Text.assemble(
            ("transform: ", "dim"),
            (f"{transform_key}", "cyan3"),
        )
    )
    space_name = (
        foreign_key_data["space"]["name"] if foreign_key_data else self.space.name
    )
    two_column_items.append(Text.assemble(("space: ", "dim"), space_name))
    branch_name = (
        foreign_key_data["branch"]["name"] if foreign_key_data else self.space.name
    )
    two_column_items.append(Text.assemble(("branch: ", "dim"), branch_name))
    # actually not name field here, but handle
    created_by_handle = (
        foreign_key_data["created_by"]["name"]
        if foreign_key_data
        else self.created_by.handle
    )
    two_column_items.append(
        Text.assemble(
            ("created_by: ", "dim"),
            (created_by_handle),
        )
    )
    two_column_items.append(
        Text.assemble(("created_at: ", "dim"), highlight_time(str(self.created_at)))
    )
    if self.n_files:
        two_column_items.append(Text.assemble(("n_files: ", "dim"), f"{self.n_files}"))
    if self.n_observations:
        two_column_items.append(
            Text.assemble(("n_observations: ", "dim"), f"{self.n_observations}")
        )
    if self.version:
        two_column_items.append(Text.assemble(("version: ", "dim"), f"{self.version}"))

    # Add two-column items in pairs
    for i in range(0, len(two_column_items), 2):
        if i + 1 < len(two_column_items):
            # Two items side by side
            left_item = two_column_items[i]
            right_item = two_column_items[i + 1]

            # Create padded version by calculating the plain text length
            left_plain_text = (
                left_item.plain if hasattr(left_item, "plain") else str(left_item)
            )
            padding_needed = max(0, 35 - len(left_plain_text))
            padding = " " * padding_needed

            general.add(Text.assemble(left_item, padding, right_item))
        else:
            # Single item (odd number)
            general.add(two_column_items[i])

    storage_root = (
        foreign_key_data["storage"]["name"] if foreign_key_data else self.storage.root
    )
    storage_key = self.key if self.key else self.uid
    if not self.key and self.overwrite_versions > 0:
        storage_key = storage_key[:-4]
    general.add(
        Text.assemble(
            ("storage path: ", "dim"),
            (storage_root, "cyan3"),
            f"/{storage_key}",
        )
    )
    return tree


def describe_collection_general(
    self: Collection,
    tree: Tree | None = None,
    foreign_key_data: dict[str, dict[str, int | str]] | None = None,
) -> Tree:
    if tree is None:
        tree = describe_header(self)

    # add general information (order is the same as in API docs)
    general = tree.add(Text("General", style="bold bright_cyan"))

    if self.key:
        general.add(Text.assemble(("key: ", "dim"), (f"{self.key}", "cyan3")))
    if self.description:
        general.add(
            Text.assemble(
                ("description: ", "dim"),
                f"{self.description}",
            )
        )

    # Two column items (short content)
    two_column_items = []

    two_column_items.append(Text.assemble(("uid: ", "dim"), f"{self.uid}"))

    transform_name = (
        foreign_key_data["transform"]["name"]
        if foreign_key_data and "transform" in foreign_key_data
        else self.transform.name
        if self.transform
        else None
    )
    if transform_name:
        two_column_items.append(
            Text.assemble(
                ("transform: ", "dim"),
                (f"{transform_name}", "cyan3"),
            )
        )

    space_name = (
        foreign_key_data["space"]["name"]
        if foreign_key_data and "space" in foreign_key_data
        else self.space.name
        if self.space
        else None
    )
    if space_name:
        two_column_items.append(Text.assemble(("space: ", "dim"), space_name))

    branch_name = (
        foreign_key_data["branch"]["name"]
        if foreign_key_data and "branch" in foreign_key_data
        else self.branch.name
        if self.branch
        else None
    )
    if branch_name:
        two_column_items.append(Text.assemble(("branch: ", "dim"), branch_name))

    created_by_handle = (
        foreign_key_data["created_by"]["name"]
        if foreign_key_data and "created_by" in foreign_key_data
        else self.created_by.handle
        if self.created_by
        else None
    )
    if created_by_handle:
        two_column_items.append(
            Text.assemble(
                ("created_by: ", "dim"),
                (created_by_handle),
            )
        )

    if self.created_at:
        two_column_items.append(
            Text.assemble(("created_at: ", "dim"), highlight_time(str(self.created_at)))
        )

    if self.version:
        two_column_items.append(Text.assemble(("version: ", "dim"), f"{self.version}"))

    # Add two-column items in pairs
    for i in range(0, len(two_column_items), 2):
        if i + 1 < len(two_column_items):
            # Two items side by side
            left_item = two_column_items[i]
            right_item = two_column_items[i + 1]

            # Create padded version by calculating the plain text length
            left_plain_text = (
                left_item.plain if hasattr(left_item, "plain") else str(left_item)
            )
            padding_needed = max(0, 45 - len(left_plain_text))
            padding = " " * padding_needed

            general.add(Text.assemble(left_item, padding, right_item))
        else:
            # Single item (odd number)
            general.add(two_column_items[i])

    return tree



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from django.db import models
from django.db.models import (
    CASCADE,
    PROTECT,
)
from lamindb_setup import _check_instance_setup

from lamindb.base.fields import (
    BooleanField,
    CharField,
    DateTimeField,
    ForeignKey,
)
from lamindb.base.users import current_user_id

from ..base.ids import base62_16
from .can_curate import CanCurate
from .sqlrecord import BaseSQLRecord, IsLink, SQLRecord

if TYPE_CHECKING:
    from datetime import datetime

    from ._feature_manager import FeatureManager
    from .artifact import Artifact
    from .block import RunBlock
    from .collection import Collection
    from .feature import FeatureValue
    from .project import Project
    from .query_set import QuerySet
    from .record import Record
    from .transform import Transform
    from .ulabel import ULabel


_TRACKING_READY: bool | None = None


def current_run() -> Run | None:
    global _TRACKING_READY

    if not _TRACKING_READY:
        _TRACKING_READY = _check_instance_setup()
    if _TRACKING_READY:
        import lamindb

        # also see get_run() in core._data
        run = lamindb._tracked.get_current_tracked_run()
        if run is None:
            run = lamindb.context.run
        return run
    else:
        return None


class TracksRun(models.Model):
    """Base class tracking latest run, creating user, and `created_at` timestamp."""

    class Meta:
        abstract = True

    created_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Time of creation of record."""
    created_by: User = ForeignKey(
        "lamindb.User",
        PROTECT,
        editable=False,
        default=current_user_id,
        related_name="+",
    )
    """Creator of record."""
    run: Run | None = ForeignKey(
        "lamindb.Run", PROTECT, null=True, default=current_run, related_name="+"
    )
    """Run that created record."""

    @overload
    def __init__(self): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class TracksUpdates(models.Model):
    """Base class tracking previous runs and `updated_at` timestamp."""

    class Meta:
        abstract = True

    updated_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Time of last update to record."""

    @overload
    def __init__(self): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class User(BaseSQLRecord, CanCurate):
    """Users.

    Every :class:`~lamindb.models.SQLRecord` has a `created_by` field that links to the creating user.

    All data in this registry is synchronized from LaminHub to ensure a universal
    user identity.

    Examples:

        Query a user by handle:

        >>> user = ln.User.get(handle="testuser1")
        >>> user
    """

    class Meta:
        app_label = "lamindb"

    _name_field: str = "handle"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(editable=False, unique=True, db_index=True, max_length=8)
    """Universal id, valid across DB instances."""
    handle: str = CharField(max_length=30, unique=True, db_index=True)
    """User handle, valid across DB instances (required)."""
    name: str | None = CharField(max_length=150, db_index=True, null=True)
    """Full name (optional)."""  # has to match hub specification, where it's also optional
    created_artifacts: Artifact
    """Artifacts created by user."""
    created_transforms: Transform
    """Transforms created by user."""
    created_runs: Run
    """Runs created by user."""
    created_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Time of creation of record."""
    updated_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Time of last update to record."""

    @overload
    def __init__(
        self,
        handle: str,
        email: str,
        name: str | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class Run(SQLRecord):
    """Runs of transforms such as the execution of a script.

    A registry to store runs of transforms, such as an executation of a script.

    Args:
        transform: `Transform` A :class:`~lamindb.Transform` record.
        name: `str | None = None` An optional name.
        params: `dict | None = None` A dictionary of parameters.
        reference: `str | None = None` For instance, an external ID or a download URL.
        reference_type: `str | None = None` For instance, `redun_id`, `nextflow_id` or `url`.
        initiated_by_run: `Run | None = None` The run that triggers this run.

    See Also:
        :func:`~lamindb.track`
            Globally track a script or notebook run.
        :func:`~lamindb.tracked`
            Track a function with this decorator.

    Examples:

        Create a run record::

            ln.Transform(key="Cell Ranger", version="7.2.0", type="pipeline").save()
            transform = ln.Transform.get(key="Cell Ranger", version="7.2.0")
            run = ln.Run(transform)

        Create a global run context for a custom transform::

            ln.track(transform=transform)
            ln.context.run  # global run

        Track a global run context for a notebook or script::

            ln.track()
            ln.context.run  # global run

        You can pass parameters to `Run(transform, params=params)` or add them later::

            run.params = {
                "learning_rate": 0.01,
                "input_dir": "s3://my-bucket/mydataset",
                "downsample": True,
                "preprocess_params": {
                    "normalization_type": "cool",
                    "subset_highlyvariable": True,
                },
            }
            run.save()

        In contrast to `.params`, features are indexed in the `Feature` registry and can reference relational categorical values.
        If you want to link feature values, use::

            run.features.add_values({
                "experiment": "My experiment 1",
            })

        Guide: :ref:`track-run-parameters`
    """

    class Meta:
        app_label = "lamindb"

    _name_field: str = "started_at"

    id: int = models.BigAutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    # default uid was changed from base62_20 to base62_16 in 1.6.0
    uid: str = CharField(
        editable=False, unique=True, db_index=True, max_length=20, default=base62_16
    )
    """Universal id, valid across DB instances."""
    name: str | None = CharField(max_length=150, null=True, db_index=True)
    """A name."""
    transform: Transform = ForeignKey("Transform", CASCADE, related_name="runs")
    """The transform :class:`~lamindb.Transform` that is being run."""
    started_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Start time of run."""
    finished_at: datetime | None = DateTimeField(db_index=True, null=True, default=None)
    """Finished time of run."""
    # we don't want to make below a OneToOne because there could be the same trivial report
    # generated for many different runs
    report: Artifact | None = ForeignKey(
        "Artifact", PROTECT, null=True, related_name="_report_of", default=None
    )
    """Report of run, e.g.. n html file."""
    _logfile: Artifact | None = ForeignKey(
        "Artifact", PROTECT, null=True, related_name="_logfile_of", default=None
    )
    """Report of run, e.g.. n html file."""
    environment: Artifact | None = ForeignKey(
        "Artifact", PROTECT, null=True, related_name="_environment_of", default=None
    )
    """Computational environment for the run.

    For instance, `Dockerfile`, `docker image`, `requirements.txt`, `environment.yml`, etc.
    """
    input_artifacts: Artifact
    """The artifacts serving as input for this run.

    Related accessor: :attr:`~lamindb.Artifact.input_of_runs`.
    """
    output_artifacts: Artifact
    """The artifacts generated by this run.

    Related accessor: via :attr:`~lamindb.Artifact.run`
    """
    input_collections: Collection
    """The collections serving as input for this run."""
    output_collections: Collection
    """The collections generated by this run."""
    input_records: Record
    """The collections serving as input for this run."""
    output_records: Record
    """The collections generated by this run."""
    params: dict = models.JSONField(null=True)
    """JSON-like parameters."""
    _feature_values: FeatureValue = models.ManyToManyField(
        "FeatureValue", through="RunFeatureValue", related_name="runs"
    )
    """Feature values."""
    reference: str | None = CharField(max_length=255, db_index=True, null=True)
    """A reference like a URL or external ID (such as from a workflow manager)."""
    reference_type: str | None = CharField(max_length=25, db_index=True, null=True)
    """Type of reference such as a workflow manager execution ID."""
    created_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Time of first creation. Mismatches ``started_at`` if the run is re-run."""
    created_by: User = ForeignKey(
        "User", CASCADE, default=current_user_id, related_name="created_runs"
    )
    """Creator of run."""
    ulabels: ULabel = models.ManyToManyField(
        "ULabel", through="RunULabel", related_name="runs"
    )
    """ULabel annotations of this transform."""
    initiated_by_run: Run | None = ForeignKey(
        "Run", CASCADE, null=True, related_name="initiated_runs", default=None
    )
    """The run that triggered the current run.

    This is not a preceding run. The preceding runs ("predecessors") is the set
    of runs that produced the output artifacts that serve as the inputs for the
    present run.
    """
    initiated_runs: Run
    """Runs that were initiated by this run."""
    projects: Project
    """Linked projects."""
    blocks: RunBlock
    """Blocks that annotate this run."""
    records: Record
    """Records that annotate this run."""
    _is_consecutive: bool | None = BooleanField(null=True)
    """Indicates whether code was consecutively executed. Is relevant for notebooks."""
    _status_code: int = models.SmallIntegerField(
        default=-3, db_default=-3, db_index=True, null=True
    )
    """Status code of the run. See the status property for mapping to string."""

    @overload
    def __init__(
        self,
        transform: Transform,
        name: str | None = None,
        params: dict | None = None,
        reference: str | None = None,
        reference_type: str | None = None,
        initiated_by_run: Run | None = None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        if len(args) == len(self._meta.concrete_fields):
            super().__init__(*args, **kwargs)
            return None
        # now we proceed with the user-facing constructor
        if len(args) > 1:
            raise ValueError("Only one non-keyword arg allowed: transform")
        transform: Transform = None
        if "transform" in kwargs or len(args) == 1:
            transform = kwargs.pop("transform") if len(args) == 0 else args[0]
        name: str | None = kwargs.pop("name", None)
        params: dict | None = kwargs.pop("params", None)
        reference: str | None = kwargs.pop("reference", None)
        reference_type: str | None = kwargs.pop("reference_type", None)
        initiated_by_run: Run | None = kwargs.pop("initiated_by_run", None)
        if transform is None:
            raise TypeError("Pass transform parameter")
        if transform._state.adding:
            raise ValueError("Please save transform record before creating a run")
        if not len(kwargs) == 0:
            raise ValueError(
                f"Only transform, name, params, reference, reference_type, initiated_by_run can be passed, but you passed: {kwargs}"
            )
        super().__init__(  # type: ignore
            transform=transform,
            name=name,
            params=params,
            reference=reference,
            reference_type=reference_type,
            initiated_by_run=initiated_by_run,
        )

    @property
    def status(self) -> str:
        """Get status of run.

        Returns the status as a string, one of: `scheduled`, `re-started`, `started`, `completed`, `errored`, or `aborted`.

        Examples:

            See the status of a run::

                run.status
                #> 'completed'

        """
        if self._status_code is None:
            return "unknown"
        status_dict = {
            -3: "scheduled",
            -2: "re-started",
            -1: "started",
            0: "completed",
            1: "errored",
            2: "aborted",
        }
        return status_dict.get(self._status_code, "unknown")

    @property
    def features(self) -> FeatureManager:
        """Manage annotations with features."""
        from ._feature_manager import FeatureManager

        return FeatureManager(self)

    @classmethod
    def filter(
        cls,
        *queries,
        **expressions,
    ) -> QuerySet:
        """Query a set of artifacts.

        Args:
            *queries: `Q` expressions.
            **expressions: Params, fields, and values passed via the Django query syntax.

        See Also:
            - Guide: :doc:`docs:registries`

        Examples:

            Query by fields::

                ln.Run.filter(key="examples/my_file.parquet")

            Query by params::

                ln.Run.filter(hyperparam_x=100)
        """
        # from Registry metaclass
        return type(cls).filter(cls, *queries, **expressions)


def delete_run_artifacts(run: Run) -> None:
    environment = None
    if run.environment is not None:
        environment = run.environment
        run.environment = None
    report = None
    if run.report is not None:
        report = run.report
        run.report = None
    if environment is not None or report is not None:
        run.save()
    if environment is not None:
        # only delete if there are no other runs attached to this environment
        if environment._environment_of.count() == 0:
            environment.delete(permanent=True)
    if report is not None:
        # only delete if there are no other runs attached to this report
        if report._report_of.count() == 0:
            report.delete(permanent=True)


class RunFeatureValue(BaseSQLRecord, IsLink):
    id: int = models.BigAutoField(primary_key=True)
    run: Run = ForeignKey(Run, CASCADE, related_name="links_featurevalue")
    # we follow the lower() case convention rather than snake case for link models
    featurevalue: FeatureValue = ForeignKey(
        "FeatureValue", PROTECT, related_name="links_run"
    )
    created_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Time of creation of record."""
    created_by: User = ForeignKey(
        "lamindb.User", PROTECT, default=current_user_id, related_name="+"
    )
    """Creator of record."""

    class Meta:
        app_label = "lamindb"
        unique_together = ("run", "featurevalue")



from __future__ import annotations

from typing import TYPE_CHECKING

import pyarrow.dataset
from lamindb_setup.core.upath import LocalPathClasses

if TYPE_CHECKING:
    from pyarrow.dataset import Dataset as PyArrowDataset
    from upath import UPath


PYARROW_SUFFIXES = (".parquet", ".csv", ".json", ".orc", ".arrow", ".feather", ".ipc")


def _open_pyarrow_dataset(paths: UPath | list[UPath], **kwargs) -> PyArrowDataset:
    if isinstance(paths, list):
        # a single path can be a directory, but a list of paths
        # has to be a flat list of files
        paths_str = []
        path0 = paths[0]
        if isinstance(path0, LocalPathClasses):
            path_to_str = lambda p: p.as_posix()
            filesystem = None
        else:
            path_to_str = lambda p: p.path
            filesystem = path0.fs
        for path in paths:
            if (
                getattr(path, "protocol", None) not in {"http", "https"}
                and path.is_dir()
            ):
                paths_str += [path_to_str(p) for p in path.rglob("*") if p.suffix != ""]
            else:
                paths_str.append(path_to_str(path))
    elif isinstance(paths, LocalPathClasses):
        paths_str, filesystem = paths.as_posix(), None
    else:
        paths_str, filesystem = paths.path, paths.fs

    return pyarrow.dataset.dataset(paths_str, filesystem=filesystem, **kwargs)



"""Examples.

.. autosummary::
   :toctree: .

   schemas
   datasets
   cellxgene
   croissant
   mlflow
   wandb

"""

from . import croissant, datasets, mlflow, schemas, wandb
from .cellxgene import _cellxgene



# Generated by Django 5.2 on 2025-01-15 14:41

import django.db.models.functions.datetime
from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0075_lamindbv1_part5"),
    ]

    operations = [
        migrations.AlterField(
            model_name="artifact",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="artifactfeaturevalue",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="artifactparamvalue",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="artifactproject",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="artifactreference",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="artifactschema",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="artifactulabel",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="collectionartifact",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="collectionproject",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="collectionreference",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="collectionulabel",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="featurevalue",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="storage",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="storage",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="transformproject",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="transformreference",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="transformulabel",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="created_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="updated_at",
            field=lamindb.base.fields.DateTimeField(
                blank=True,
                db_default=django.db.models.functions.datetime.Now(),
                db_index=True,
                editable=False,
            ),
        ),
        migrations.AlterField(
            model_name="artifactfeaturevalue",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="artifactparamvalue",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="artifactproject",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="artifactreference",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="artifactschema",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="artifactulabel",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="collectionartifact",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="collectionproject",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="collectionreference",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="collectionulabel",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="feature",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="featurevalue",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="param",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="schema",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="storage",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="transformproject",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="transformreference",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="transformulabel",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
        migrations.AlterField(
            model_name="ulabel",
            name="created_by",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=lamindb.base.users.current_user_id,
                editable=False,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.user",
            ),
        ),
    ]



import bionty as bt
import lamindb as ln
import pytest


@pytest.mark.parametrize("permanent", [True, False])
def test_delete_qs(permanent):
    """Test deletion behavior for small (1) and large (>=2) querysets.

    Small querysets delete individually, large ones trigger bulk delete."""
    ln.settings.creation.search_names = False
    labels = [ln.Record(name=f"label_{i}") for i in range(3)]
    ln.settings.creation.search_names = True
    ln.save(labels)
    ln.Record.filter(name__startswith="label_").delete(permanent=permanent)
    assert ln.Record.filter(name__startswith="label_", branch_id=-1).count() == (
        0 if permanent else 3
    )
    assert ln.ULabel.filter(name__startswith="label_").count() == 0


def test_recreate_soft_deleted_record():
    # testing soft delete and recreate with postgres (sqlite is tested in curators/test_records.py)
    # soft delete a record, then recreate it with some changes
    record = bt.Ethnicity.from_source(ontology_id="HANCESTRO:0006").save()
    assert record.branch_id == 1
    record.delete()
    assert record.branch_id == -1
    # now recreate the same record from ontology_id with a different description
    # there's a unique constraint on ontology_id, so this should recover the trashed record
    record = bt.Ethnicity.from_source(ontology_id="HANCESTRO:0006")
    record.description = "new description"
    record.save()
    # now this record is recovered from the trash with the new description
    assert record.branch_id == 1
    assert record.description == "new description"
    bt.Ethnicity.objects.filter().delete()



# Generated by Django 5.2 on 2025-10-15 09:35

from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0134_run_params"),
    ]

    operations = [
        migrations.AlterField(
            model_name="collection",
            name="hash",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=22, null=True
            ),
        ),
        migrations.AddConstraint(
            model_name="collection",
            constraint=models.UniqueConstraint(
                fields=("key", "hash"), name="unique_collection_key_hash_not_null"
            ),
        ),
    ]



import lamindb as ln

ln.examples.datasets.mini_immuno.define_features_labels()
adata = ln.examples.datasets.mini_immuno.get_dataset1(otype="AnnData")
schema = ln.Schema.get(name="Study metadata schema")
artifact = ln.Artifact.from_anndata(
    adata, schema=schema, key="examples/mini_immuno_uns.h5ad"
)
artifact.describe()



import lamindb as ln  # noqa
import hubmodule
import hubmodule.models as hm
from uuid import uuid4
from hubmodule._migrate import _apply_migrations_with_tracking
from hubmodule._setup import _setup_extensions, _setup_secret, _setup_utils_jwt
from hubmodule._rls import RLSGenerator
from laminhub_rest.core.postgres import DbRoleHandler
from pathlib import Path


# create a db connection url that works with RLS
instance_id = ln.setup.settings.instance._id


def create_jwt_user(dsn_admin: str, jwt_role_name: str):
    db_role_handler = DbRoleHandler(dsn_admin)
    jwt_db_url = db_role_handler.create(
        jwt_role_name, expires_in=None, alter_if_exists=True
    )
    db_role_handler.permission.grant_write_jwt(jwt_role_name)
    return jwt_db_url


pgurl = "postgresql://postgres:pwd@0.0.0.0:5432/pgtest"  # admin db connection url
jwt_role_name = f"{instance_id.hex}_jwt"
jwt_db_url = create_jwt_user(pgurl, jwt_role_name=jwt_role_name)

_setup_extensions(pgurl)
_setup_secret(pgurl)
_setup_utils_jwt(pgurl)
migrations_sql_dir = Path(hubmodule.__file__).parent / "sql/0004_migrations"
_apply_migrations_with_tracking(pgurl, migrations_sql_dir)

rls_generator = RLSGenerator(pgurl, jwt_role_name=jwt_role_name, public_role_name=None)

for i, table in enumerate(rls_generator._list_tables()):
    print(i, table.table_name, table.foreign_keys, table.has_space_id)

rls_generator.setup()


print("Created jwt db connection")

# create models

full_access = ln.Space(name="full access").save()  # type: ignore
select_access = ln.Space(name="select access").save()  # type: ignore
no_access = ln.Space(name="no access").save()  # type: ignore
# set read role for the default space
account = hm.Account(
    id=ln.setup.settings.user._uuid.hex, uid="accntid1", role="read"
).save()

# no access space
ulabel = ln.ULabel(name="no_access_ulabel")
ulabel.space = no_access
ulabel.save()
# set up access to this individual record with a dummy role,
# will work only after the role is changed to read, write or admin
hm.AccessRecord(
    account=account, record_type="lamindb_ulabel", record_id=ulabel.id, role="dummy"
).save()

project = ln.Project(name="No_access_project")  # type: ignore
project.space = no_access
project.save()

hm.AccessRecord(
    account=account, record_type="lamindb_project", record_id=project.id, role="dummy"
).save()

# setup write access space
hm.AccessSpace(account=account, space=full_access, role="write").save()

ulabel = ln.ULabel(name="full_access_ulabel")
ulabel.space = full_access
ulabel.save()
# setup read access space
hm.AccessSpace(account=account, space=select_access, role="read").save()

ulabel = ln.ULabel(name="select_ulabel")
ulabel.space = select_access
ulabel.save()
# artificial but better to test
# create a link table referencing rows in different spaces
ulabel.projects.add(project)

# default space, only select access by default
ulabel = ln.ULabel(name="default_space_ulabel").save()
ulabel.projects.add(project)

project = ln.Project(name="default_space_project").save()
ulabel.projects.add(project)

# create a link table referencing ulabel from the default space and project from select space
project = ln.Project(name="select_project")
project.space = select_access
project.save()

ulabel.projects.add(project)

# setup team and relevent models
team_access = ln.Space(name="team access").save()  # type: ignore
team = hm.Team(id=uuid4().hex, uid="teamuiduid11", name="test_team", role="read").save()
hm.AccountTeam(account=account, team=team).save()
hm.AccessSpace(team=team, space=team_access, role="read").save()

feature = ln.Feature(name="team_access_feature", dtype=float)
feature.space = team_access
feature.save()

# artifact for testing tracking error and artifactblock
artifact = ln.Artifact("README.md", description="test tracking error")
artifact.space = select_access
artifact.save()

# artifact for testing tracking error and locking
artifact = ln.Artifact(".gitignore", description="test locking")
artifact.space = full_access
artifact.is_locked = True
artifact.save()

print("Created models")

# save jwt db connection

ln.setup.settings.instance._db = jwt_db_url
ln.setup.settings.instance._persist()



# Generated by Django 5.2 on 2025-06-03 19:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0104_alter_branch_uid"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="record",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_type", True)),
                fields=("name",),
                name="unique_name",
            ),
        ),
    ]



# Generated by Django 5.1.12 on 2025-10-02 10:27

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0129_remove_feature_page_remove_project_page_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="branch",
            name="space",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                db_default=1,
                default=1,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="lamindb.space",
            ),
        ),
        migrations.AlterField(
            model_name="artifactblock",
            name="artifact",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.artifact",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="artifactblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="branchblock",
            name="branch",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.branch",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="branchblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="collectionblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="featureblock",
            name="feature",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.feature",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="featureblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="projectblock",
            name="project",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.project",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="projectblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="recordblock",
            name="record",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.record",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="recordblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="rootblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="runblock",
            name="run",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.run",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="runblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="schemablock",
            name="schema",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.schema",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="schemablock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="spaceblock",
            name="space",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="blocks",
                to="lamindb.space",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="spaceblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
        migrations.AlterField(
            model_name="transformblock",
            name="vertical_pos",
            field=models.FloatField(db_default=0, db_index=True, default=0),
        ),
    ]



"""Universal IDs.

Base generators:

.. autosummary::
   :toctree: .

   base26
   base62
   base64

`uid` generators:

.. autosummary::
   :toctree: .

   base62_8
   base62_12
   base62_16
   base62_20


Collision probabilities
=======================

8 base62 characters (`62**8=2e+14`):

======= ===========
n       p_collision
======= ===========
100k    2e-05
1M      2e-03
======= ===========

12 base62 characters (`62**12=3e+21`):

======= ===========
n       p_collision
======= ===========
100M    2e-06
1B      2e-04
======= ===========

16 base62 characters (`62**16=5e+28`):

======= ===========
n       p_collision
======= ===========
1e12    7e-05
1e13    7e-03
======= ===========

20 base62 characters (`62**20=7e+35`) roughly matches UUID (`2**122=5e+36`):

======= ===========
n       p_collision
======= ===========
1e16    7e-05
1e17    7e-03
======= ===========

See `source <https://lamin.ai/laminlabs/lamindata/transform/t2xCdMB9v5wL>`__.

"""

import secrets
import string


def base64(n_char: int) -> str:
    """Random Base64 string."""
    alphabet = string.digits + string.ascii_letters.swapcase() + "_" + "-"
    uid = "".join(secrets.choice(alphabet) for i in range(n_char))
    return uid


def base62(n_char: int) -> str:
    """Random Base62 string."""
    alphabet = string.digits + string.ascii_letters.swapcase()
    uid = "".join(secrets.choice(alphabet) for i in range(n_char))
    return uid


def base26(n_char: int):
    """ASCII lowercase."""
    alphabet = string.ascii_lowercase
    uid = "".join(secrets.choice(alphabet) for i in range(n_char))
    return uid


def base62_4() -> str:
    return base62(4)


def base62_8() -> str:
    """Random Base62 string of length 8."""
    return base62(8)


def base62_12() -> str:
    """Random Base62 string of length 12."""
    return base62(12)


def base62_16() -> str:
    """Random Base62 string of length 16."""
    return base62(16)


def base62_20() -> str:
    """Random Base62 string of length 20."""
    return base62(20)


def base62_24() -> str:
    """Random Base62 string of length 24."""
    return base62(24)



# Generated by Django 5.1.12 on 2025-09-30 17:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "lamindb",
            "0126_alter_artifact_is_locked_alter_collection_is_locked_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="run",
            name="_status_code",
            field=models.SmallIntegerField(
                db_default=-3, db_index=True, default=-3, null=True
            ),
        ),
        migrations.AddConstraint(
            model_name="feature",
            constraint=models.CheckConstraint(
                condition=models.Q(
                    ("is_type", True), ("dtype__isnull", False), _connector="OR"
                ),
                name="dtype_not_null_when_is_type_false",
            ),
        ),
    ]



import lamindb as ln

from .define_schema_df_metadata import study_metadata_schema

df = ln.examples.datasets.mini_immuno.get_dataset1(otype="DataFrame")
schema = ln.Schema(
    features=[ln.Feature(name="perturbation", dtype="str").save()],
    slots={"attrs": study_metadata_schema},
    otype="DataFrame",
).save()
curator = ln.curators.DataFrameCurator(df, schema=schema)
curator.validate()
artifact = curator.save_artifact(key="examples/df_with_attrs.parquet")
artifact.describe()



import argparse
import lamindb as ln


@ln.tracked()
def subset_dataframe(
    artifact: ln.Artifact,
    subset_rows: int = 2,
    subset_cols: int = 2,
    run: ln.Run | None = None,
) -> ln.Artifact:
    dataset = artifact.load(is_run_input=run)
    new_data = dataset.iloc[:subset_rows, :subset_cols]
    new_key = artifact.key.replace(".parquet", "_subsetted.parquet")
    return ln.Artifact.from_dataframe(new_data, key=new_key, run=run).save()


if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument("--subset", action="store_true")
    args = p.parse_args()

    params = {"is_subset": args.subset}

    ln.track(params=params)

    if args.subset:
        df = ln.examples.datasets.mini_immuno.get_dataset1(otype="DataFrame")
        artifact = ln.Artifact.from_dataframe(
            df, key="my_analysis/dataset.parquet"
        ).save()
        subsetted_artifact = subset_dataframe(artifact)

    ln.finish()



import os
import shutil
from pathlib import Path

import nox
from laminci import upload_docs_artifact
from laminci.nox import (
    build_docs,
    login_testuser1,
    login_testuser2,
    run,
    run_pre_commit,
)

# we'd like to aggregate coverage information across sessions
# and for this the code needs to be located in the same
# directory in every github action runner
# this also allows to break out an installation section
nox.options.default_venv_backend = "none"

IS_PR = os.getenv("GITHUB_EVENT_NAME") != "push"
CI = os.environ.get("CI")


GROUPS = {}
GROUPS["tutorial"] = [
    "transfer.ipynb",
    "arrays.ipynb",
    "registries.ipynb",
]
GROUPS["guide"] = [
    "track.ipynb",
    "curate.ipynb",
]
GROUPS["biology"] = [
    "manage-ontologies.ipynb",
]


@nox.session
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session
def install(session):
    base_deps = [
        "./sub/lamin-cli",
        "./sub/lamindb-setup",
        "./sub/bionty",
    ]
    top_deps = [
        ".[dev]",
    ]
    cmds = [
        f"uv pip install {'--system' if CI else ''} --no-cache-dir {' '.join(base_deps)}",
    ] + [
        f"uv pip install {'--system' if CI else ''} --no-cache-dir -e {dep}"
        for dep in top_deps
    ]
    [run(session, line) for line in cmds]


@nox.session
@nox.parametrize(
    "group",
    [
        "unit-core",
        "unit-storage",
        "tutorial",
        "guide",
        "biology",
        "faq",
        "storage",
        "curator",
        "integrations",
        "docs",
        "cli",
        "permissions",
    ],
)
def install_ci(session, group):
    extras = ""
    if group == "unit-core":
        extras += "zarr,fcs"
        # tiledbsoma dependency, specifying it here explicitly
        # otherwise there are problems with uv resolver
        run(session, "uv pip install --system scanpy")
        run(session, "uv pip install --system tiledbsoma")  # test TiledbsomaCatManager
        run(session, "uv pip install --system mudata")
        # spatialdata dependency, specifying it here explicitly
        # otherwise there are problems with uv resolver
        run(session, "uv pip install --system xarray-dataclasses")
        run(session, "uv pip install --system spatialdata")
    elif group == "unit-storage":
        extras += "zarr,gcp"
        run(session, "uv pip install --system huggingface_hub")
        # tiledbsoma dependency, specifying it here explicitly
        # otherwise there are problems with uv resolver
        run(session, "uv pip install --system scanpy")
        run(session, "uv pip install --system tiledbsoma")
        run(session, "uv pip install --system polars")
    elif group == "tutorial":
        # anndata here to prevent installing older version on release
        run(session, "uv pip install --system huggingface_hub polars anndata==0.12.1")
    elif group == "guide":
        extras += "zarr"
        run(session, "uv pip install --system scanpy mudata spatialdata tiledbsoma")
    elif group == "biology":
        extras += "fcs"
        run(session, "uv pip install --system ipywidgets")
    elif group == "faq":
        extras += "zarr"
    elif group == "storage":
        extras += "zarr"
        run(
            session,
            "uv pip install --system --no-deps ./sub/wetlab",
        )
        run(session, "uv pip install --system vitessce")
    elif group == "curator":
        extras += "zarr"
        run(
            session,
            "uv pip install --system --no-deps ./sub/wetlab",
        )
        # spatialdata dependency, specifying it here explicitly
        # otherwise there are problems with uv resolver
        run(session, "uv pip install --system xarray-dataclasses")
        run(
            session,
            "uv pip install --system spatialdata",
        )
        run(session, "uv pip install --system tiledbsoma")
    elif group == "integrations":
        run(session, "uv pip install --system lightning")
    elif group == "docs":
        extras += "zarr"
        # spatialdata dependency, specifying it here explicitly
        # otherwise there are problems with uv resolver
        run(session, "uv pip install --system xarray-dataclasses")
        run(
            session,
            "uv pip install --system mudata spatialdata",
        )
        run(
            session,
            "uv pip install --system --no-deps ./sub/wetlab",
        )
    elif group == "cli":
        pass
    elif group == "permissions":
        run(
            session,
            "uv pip install --system --no-deps ./laminhub/rest-hub/laminhub_rest/hubmodule",
        )
        # check that just installing psycopg (psycopg3) doesn't break fine-grained access
        # comment out for now, this is also tested in lamindb-setup hub-local
        # run(session, "uv pip install --system psycopg[binary]")

    extras = "," + extras if extras != "" else extras
    run(session, f"uv pip install --system -e .[dev{extras}]")

    if group == "permissions":
        # have to install after lamindb installation
        # because lamindb downgrades django
        run(
            session,
            "uv pip install --system sentry_sdk line_profiler setuptools wheel==0.45.1 flit",
        )
        run(
            session,
            "uv pip install --system -e ./laminhub/rest-hub --no-build-isolation",
        )
    # needed here till the next release of lamindb-setup
    run(session, "uv pip install --system httpx_retries")
    # on the release branch, do not use submodules but run with pypi install
    # only exception is the docs group which should always use the submodule
    # to push docs fixes fast
    # installing this after lamindb to be sure that these packages won't be reinstaled
    # during lamindb installation
    if IS_PR or group == "docs":
        run(
            session,
            "uv pip install --system --no-deps ./sub/lamindb-setup ./sub/lamin-cli ./sub/bionty",
        )


@nox.session
def configure_coverage(session) -> None:
    """Write a coverage config file, adding extra patterns to omit."""
    import tomlkit

    groups_str = session.posargs[0]  # first positional argument

    print(groups_str)  # for debugging
    # so that we don't change this away from string
    assert isinstance(groups_str, str)  # noqa: S101

    if "curator" not in groups_str:
        extra_omit_patterns = ["**/curators/*"]
    else:
        extra_omit_patterns = []

    # Read patterns from pyproject.toml
    base_config_path = Path("pyproject.toml")
    with open(base_config_path) as f:
        config = tomlkit.load(f)

    # Update the omit patterns
    base_patterns = config["tool"]["coverage"]["run"]["omit"]
    all_patterns = base_patterns + extra_omit_patterns
    config["tool"]["coverage"]["run"]["omit"] = all_patterns

    # Write back to pyproject.toml
    with open(base_config_path, "w") as f:
        tomlkit.dump(config, f)

    print(base_config_path.read_text())


@nox.session
@nox.parametrize(
    "group",
    [
        "unit-core",
        "unit-storage",
        "curator",
        "integrations",
        "tutorial",
        "guide",
        "biology",
        "faq",
        "storage",
        "cli",
        "permissions",
    ],
)
def test(session, group):
    login_testuser2(session)
    login_testuser1(session)
    run(session, "lamin settings set private-django-api true")
    coverage_args = "--cov=lamindb --cov-config=pyproject.toml --cov-append --cov-report=term-missing"
    duration_args = "--durations=10"
    if group == "unit-core":
        run(
            session,
            f"pytest {coverage_args} ./tests/core {duration_args}",
        )
    elif group == "unit-storage":
        login_testuser2(session)  # shouldn't be necessary but is for now
        run(session, f"pytest -s {coverage_args} ./tests/storage {duration_args}")
    elif group == "tutorial":
        run(session, "lamin logout")
        run(
            session, f"pytest -s {coverage_args} ./docs/test_notebooks.py::test_{group}"
        )
    elif group == "guide":
        run(
            session,
            f"pytest -s {coverage_args} ./docs/test_notebooks.py::test_{group}",
        )
    elif group == "biology":
        run(
            session,
            f"pytest -s {coverage_args} ./docs/test_notebooks.py::test_{group}",
        )
    elif group == "faq":
        run(session, f"pytest -s {coverage_args} ./docs/faq")
    elif group == "storage":
        run(session, f"pytest -s {coverage_args} ./docs/storage")
    elif group == "curator":
        run(
            session,
            f"pytest {coverage_args} tests/curators {duration_args}",
        )
    elif group == "integrations":
        run(session, f"pytest -s {coverage_args} tests/integrations")
    elif group == "cli":
        run(
            session,
            f"pytest {coverage_args} ./sub/lamin-cli/tests/core {duration_args}",
        )
    elif group == "permissions":
        run(session, f"pytest {coverage_args} ./tests/permissions")
    # move artifacts into right place
    if group in {"tutorial", "guide", "biology"}:
        target_dir = Path(f"./docs/{group}")
        target_dir.mkdir(exist_ok=True)
        for filename in GROUPS[group]:
            shutil.copy(Path("docs") / filename, target_dir / filename)


@nox.session
def clidocs(session):
    def generate_cli_docs():
        os.environ["NO_RICH"] = "1"
        from lamin_cli.__main__ import COMMAND_GROUPS, _generate_help

        page = "# CLI\n\n"
        helps = _generate_help()

        # First, add the main lamin command
        main_help = helps.get("main")
        if main_help:
            help_string = main_help["help"].replace("Usage: main", "Usage: lamin")
            help_docstring = main_help["docstring"]
            if help_docstring:
                page += f"{help_docstring}\n\n"
            # below is ugly
            # page += f"```text\n{help_string}\n```\n\n"

        # Create a mapping of command names to their full keys in helps
        command_to_key = {}
        for name in helps.keys():
            names = name.split(" ")
            if len(names) == 2:  # e.g., "lamin connect"
                command_name = names[1]
                command_to_key[command_name] = name

        # Group commands by their categories
        command_groups = COMMAND_GROUPS.get("lamin", [])
        processed_commands = set()

        for group in command_groups:
            group_name = group["name"]
            group_commands = group["commands"]

            page += f"## {group_name}\n\n"

            for command_name in group_commands:
                if command_name in command_to_key:
                    full_key = command_to_key[command_name]
                    help_dict = helps[full_key]
                    processed_commands.add(command_name)

                    help_string = help_dict["help"].replace("Usage: main", "lamin")
                    help_docstring = help_dict["docstring"]

                    page += f"### {command_name}\n\n"
                    if help_docstring:
                        page += f"{help_docstring}\n\n"
                    page += f"Usage:\n```text\n{help_string}\n```\n\n"

        # Add any remaining commands that aren't in groups
        remaining_commands = []
        for command_name, full_key in command_to_key.items():
            if command_name not in processed_commands:
                remaining_commands.append((command_name, full_key))

        if remaining_commands:
            page += "## Other\n\n"
            for command_name, full_key in remaining_commands:
                help_dict = helps[full_key]
                help_string = help_dict["help"].replace("Usage: main", "Usage: lamin")
                help_docstring = help_dict["docstring"]

                page += f"### lamin {command_name}\n\n"
                if help_docstring:
                    page += f"{help_docstring}\n\n"
                page += f"```text\n{help_string}\n```\n\n"

        Path("./docs/cli.md").write_text(page)

    generate_cli_docs()


@nox.session
def docs(session):
    # move artifacts into right place
    for group in ["tutorial", "guide", "biology", "faq", "storage"]:
        if Path(f"./docs-{group}").exists():
            if Path(f"./docs/{group}").exists():
                shutil.rmtree(f"./docs/{group}")
            Path(f"./docs-{group}").rename(f"./docs/{group}")
        # move back to root level
        if group in {"tutorial", "guide", "biology"}:
            for path in Path(f"./docs/{group}").glob("*"):
                path.rename(f"./docs/{path.name}")
    run(
        session,
        "lamin init --storage ./docsbuild --modules bionty,wetlab",
    )
    build_docs(session, strip_prefix=True, strict=False)
    upload_docs_artifact(aws=True)



import lamindb as ln

schema = ln.examples.datasets.mini_immuno.define_mini_immuno_schema_flexible()
df = ln.examples.datasets.mini_immuno.get_dataset1(otype="DataFrame")
df.pop("donor")  # remove donor column to trigger validation error
try:
    artifact = ln.Artifact.from_dataframe(
        df, key="examples/dataset1.parquet", schema=schema
    ).save()
except ln.errors.ValidationError as error:
    print(error)



import subprocess
import time
from pathlib import Path
from uuid import uuid4

import hubmodule.models as hm
import lamindb as ln
import psycopg2
import pytest
from django.db import connection, transaction
from django.db.utils import IntegrityError, InternalError, ProgrammingError
from jwt_utils import sign_jwt
from lamindb.models.artifact import track_run_input
from lamindb_setup.core.django import DBToken, db_token_manager
from psycopg2.extensions import adapt

pgurl = "postgresql://postgres:pwd@0.0.0.0:5432/pgtest"  # admin db connection url

user_uuid = ln.setup.settings.user._uuid.hex
expiration = time.time() + 2000
# full collaborator token
token = sign_jwt(
    pgurl, {"account_id": user_uuid, "exp": expiration, "type": "collaborator"}
)
# read-only token
token_read = sign_jwt(
    pgurl, {"account_id": user_uuid, "exp": expiration, "type": "read-only"}
)
# init an instance of DBToken manually
db_token = DBToken({})
db_token._token = token
db_token._token_query = f"SELECT set_token({adapt(token).getquoted().decode()}, true);"
db_token._expiration = expiration

db_token_manager.set(db_token)


def test_authentication():
    # just check that the token was setup
    with connection.cursor() as cur:
        cur.execute(
            "SELECT 1 in (SELECT id FROM public.check_access() WHERE role = 'read');"
        )
        result = cur.fetchall()[0][0]
    assert result
    # check querying without setting jwt
    with (
        pytest.raises(psycopg2.errors.RaiseException),
        connection.connection.cursor() as cur,
    ):
        cur.execute("SELECT * FROM lamindb_ulabel;")
    # test that auth can't be hijacked
    # false table created before
    with (
        pytest.raises(psycopg2.errors.DuplicateTable),
        connection.connection.cursor() as cur,
    ):
        cur.execute(
            """
            CREATE TEMP TABLE access(
                id int,
                role varchar(20),
                type text
            ) ON COMMIT DROP;
            SELECT set_token(%s);
            """,
            (token,),
        )
    # check that jwt user can't set arbitrary account_id manually
    with (
        pytest.raises(psycopg2.errors.RaiseException),
        connection.connection.cursor() as cur,
    ):
        cur.execute(
            """
            CREATE TEMP TABLE access(
                id int,
                role varchar(20),
                type text
            ) ON COMMIT DROP;
            INSERT INTO access (id, role, type)
            VALUES (1, 'admin', 'space');
            SELECT * FROM check_access();
            """
        )
    # check manual insert
    with (
        pytest.raises(psycopg2.errors.InsufficientPrivilege),
        connection.connection.cursor() as cur,
    ):
        cur.execute(
            """
            SELECT set_token(%s);
            INSERT INTO access (id, role, type)
            VALUES (1, 'admin', 'space');
            """,
            (token,),
        )
    # test access to the security schema
    with (
        pytest.raises(psycopg2.errors.InsufficientPrivilege),
        connection.connection.cursor() as cur,
    ):
        cur.execute("SELECT security.get_secret('jwt_secret');")
    # test read-only token
    with connection.connection.cursor() as cur:
        cur.execute("SELECT set_token(%s); SELECT * FROM check_access()", (token_read,))
        result = cur.fetchall()
    assert len(result) == 1
    assert result[0] == (1, "read", "space")


def test_fine_grained_permissions_account():
    # check select
    assert ln.ULabel.filter().count() == 3
    assert ln.Project.filter().count() == 2

    ulabel = ln.ULabel.get(name="default_space_ulabel")
    assert ulabel.projects.all().count() == 2
    # check delete
    # should delete
    ln.ULabel.get(name="full_access_ulabel").delete(permanent=True)
    assert ln.ULabel.filter().count() == 2
    # should not delete, does not error for some reason
    ln.ULabel.get(name="select_ulabel").delete(permanent=True)
    assert ln.ULabel.filter().count() == 2
    # default space
    ulabel.delete(permanent=True)
    assert ln.ULabel.filter().count() == 2
    # check insert
    # should succeed
    space = ln.Space.get(name="full access")
    ulabel = ln.ULabel(name="new label")
    ulabel.space = space
    ulabel.save()
    # should fail
    with pytest.raises(ln.errors.NoWriteAccess):
        ln.ULabel(name="new label fail").save()
    for space_name in ["select access", "no access"]:
        space = ln.Space.get(name=space_name)
        ulabel = ln.ULabel(name="new label fail")
        ulabel.space = space
        with pytest.raises(ln.errors.NoWriteAccess):
            ulabel.save()
    # check update
    # should succeed
    ulabel = ln.ULabel.get(name="new label")
    ulabel.name = "new label update"
    ulabel.save()
    ulabel = ln.ULabel.get(name="new label update")  # check that it is saved
    # should fail
    ulabel = ln.ULabel.get(name="select_ulabel")
    ulabel.name = "select_ulabel update"
    with pytest.raises(ln.errors.NoWriteAccess):
        ulabel.save()
    # default space
    ulabel = ln.ULabel.get(name="default_space_ulabel")
    ulabel.name = "default_space_ulabel update"
    with pytest.raises(ln.errors.NoWriteAccess):
        ulabel.save()
    # check link tables
    # check insert
    project = ln.Project(name="Myproject")
    project.space = ln.Space.get(name="full access")
    project.save()
    ulabel = ln.ULabel.get(name="new label update")
    ulabel.projects.add(project)
    assert ulabel.projects.all().count() == 1
    # check select of a link table referencing unavailable rows
    assert ln.ULabel.get(name="select_ulabel").projects.all().count() == 0
    # test RootBlock, can do due to write access to some spaces
    root_block = ln.models.RootBlock(name="instance", content="test").save()
    root_block.content = "test 2"
    root_block.save()
    # test SpaceBlock
    space = ln.Space.get(name="select access")
    with pytest.raises(ln.errors.NoWriteAccess):
        ln.models.SpaceBlock(space=space, content="test").save()
    # test ArtifactBlock, artifact is read-only
    artifact = ln.Artifact.get(description="test tracking error")
    with pytest.raises(ProgrammingError):
        ln.models.ArtifactBlock(artifact=artifact, content="test").save()
    # test BranchBlock, the account is read-only
    branch = ln.Branch.get(1)  # main branch in all space
    with pytest.raises(ProgrammingError):
        ln.models.BranchBlock(branch=branch, content="test").save()


def test_fine_grained_permissions_team():
    assert ln.Feature.filter().count() == 1
    ln.Feature.get(name="team_access_feature")


def test_fine_grained_permissions_single_records():
    assert not ln.ULabel.filter(name="no_access_ulabel").exists()
    assert not ln.Project.filter(name="No_access_project").exists()

    # switch access to this ulabel to read
    with psycopg2.connect(pgurl) as conn, conn.cursor() as cur:
        cur.execute(
            """
            UPDATE hubmodule_accessrecord SET role = 'read'
            WHERE account_id = %s AND record_type = 'lamindb_ulabel'
            """,
            (user_uuid,),
        )

    ulabel = ln.ULabel.get(name="no_access_ulabel")

    new_name = "new_name_single_rls_access_ulabel"
    ulabel.name = new_name
    with pytest.raises(ln.errors.NoWriteAccess):
        ulabel.save()

    # switch access to this ulabel to write
    with psycopg2.connect(pgurl) as conn, conn.cursor() as cur:
        cur.execute(
            """
            UPDATE hubmodule_accessrecord SET role = 'write'
            WHERE account_id = %s AND record_type = 'lamindb_ulabel'
            """,
            (user_uuid,),
        )

    ulabel.save()

    # switch access to this ulabel to write
    with psycopg2.connect(pgurl) as conn, conn.cursor() as cur:
        cur.execute(
            """
            UPDATE hubmodule_accessrecord SET role = 'read'
            WHERE account_id = %s AND record_type = 'lamindb_project'
            """,
            (user_uuid,),
        )

    project = ln.Project.get(name="No_access_project")
    # can't insert into lamindb_ulabelproject because the project is still read-only
    with pytest.raises(ProgrammingError):
        ulabel.projects.add(project)

    with psycopg2.connect(pgurl) as conn, conn.cursor() as cur:
        cur.execute(
            """
            UPDATE hubmodule_accessrecord SET role = 'write'
            WHERE account_id = %s AND record_type = 'lamindb_project'
            """,
            (user_uuid,),
        )

    ulabel.projects.add(project)
    assert ulabel.projects.count() == 1

    ulabel.delete(permanent=True)
    assert not ln.ULabel.filter(name="no_access_ulabel").exists()


# tests that token is set properly in atomic blocks
def test_atomic():
    with transaction.atomic():
        assert ln.Feature.filter().count() == 1
        # test with nested
        with transaction.atomic():
            assert ln.Feature.filter().count() == 1

            feature = ln.Feature(name="atomic_feature", dtype=float)
            feature.space = ln.Space.get(name="full access")
            feature.save()

    assert ln.Feature.filter().count() == 2


def test_utility_tables():
    # can select in these tables
    assert ln.User.filter().count() == 1
    assert ln.Space.filter().count() == 5
    # can't select
    assert hm.Account.filter().count() == 0
    assert hm.Team.filter().count() == 0
    assert hm.AccountTeam.filter().count() == 0
    assert hm.AccessSpace.filter().count() == 0
    assert hm.AccessRecord.filter().count() == 0
    # can't update a space
    space = ln.Space.get(id=1)  # default space
    space.name = "new name"
    with pytest.raises(ProgrammingError):
        space.save()
    # can't update a user
    user = ln.User.filter().one()
    user.name = "new name"
    # as we allow insert but not update on the user table
    # it looks like the db raises IntegrityError insead of the rls error
    # because just tries to insert with the same id and fails
    with pytest.raises(IntegrityError):
        user.save()
    # can insert a user because has write access to a space
    ln.User(handle="insert_new_user", uid="someuidd").save()
    assert ln.User.filter().count() == 2
    # can't insert
    with pytest.raises(ProgrammingError):
        ln.Space(name="new space").save()

    with pytest.raises(ProgrammingError):
        hm.Account(id=uuid4().hex, uid="accntid2", role="admin").save()


def test_write_role():
    # switch user role to write
    with psycopg2.connect(pgurl) as conn, conn.cursor() as cur:
        cur.execute(
            "UPDATE hubmodule_account SET role = 'write' WHERE id = %s", (user_uuid,)
        )

    ln.ULabel(name="new label account default space").save()

    # switch user role back to read and team role to write
    with psycopg2.connect(pgurl) as conn, conn.cursor() as cur:
        cur.execute(
            "UPDATE hubmodule_account SET role = 'read' WHERE id = %s", (user_uuid,)
        )
        cur.execute(
            "UPDATE hubmodule_team SET role = 'write' WHERE uid = 'teamuiduid11'",
        )

    ln.ULabel(name="new label team default space").save()

    with psycopg2.connect(pgurl) as conn, conn.cursor() as cur:
        cur.execute(
            "UPDATE hubmodule_team SET role = 'read' WHERE uid = 'teamuiduid11'",
        )


def test_locking():
    artifact = ln.Artifact.get(description="test locking")
    artifact.description = "new description"
    with pytest.raises(ln.errors.NoWriteAccess) as e:
        artifact.save()
    assert "It is not allowed to modify or create locked" in str(e)


def test_tracking_error():
    # switch user role to write to create the transform and run
    with psycopg2.connect(pgurl) as conn, conn.cursor() as cur:
        cur.execute(
            "UPDATE hubmodule_account SET role = 'write' WHERE id = %s", (user_uuid,)
        )

    artifact = ln.Artifact.get(description="test tracking error")

    transform = ln.Transform(key="My transform").save()
    run = ln.Run(transform).save()

    # this error because ln.setup.settings.instance._db_permissions is not jwt
    # it is None
    with pytest.raises(ln.errors.NoWriteAccess) as e:
        track_run_input(artifact, run)
    assert "You’re not allowed to write to the instance " in str(e)

    # the instance is local so we set this manually
    ln.setup.settings.instance._db_permissions = "jwt"
    # artifact.space is not available for writes
    with pytest.raises(ln.errors.NoWriteAccess) as e:
        track_run_input(artifact, run)
    assert "You’re not allowed to write to the space " in str(e)

    # this artifact is locked
    artifact = ln.Artifact.get(description="test locking")
    with pytest.raises(ln.errors.NoWriteAccess) as e:
        track_run_input(artifact, run)
    assert "It is not allowed to modify locked records" in str(e)

    # switch user role back to read
    with psycopg2.connect(pgurl) as conn, conn.cursor() as cur:
        cur.execute(
            "UPDATE hubmodule_account SET role = 'read' WHERE id = %s", (user_uuid,)
        )
    # as the user is read-only now, 2 spaces are unavailable for writes (artifact.space, run.space)
    artifact = ln.Artifact.get(description="test tracking error")
    with pytest.raises(ln.errors.NoWriteAccess) as e:
        track_run_input(artifact, run)
    assert "You’re not allowed to write to the spaces " in str(e)

    ln.setup.settings.instance._db_permissions = None


def test_token_reset():
    db_token_manager.reset()

    # account_id is not set
    with pytest.raises(InternalError) as error:
        ln.ULabel.filter().count()
    assert "JWT is not set" in error.exconly()

    with pytest.raises(InternalError) as error, transaction.atomic():
        ln.ULabel.filter().count()
    assert "JWT is not set" in error.exconly()


def test_lamin_dev():
    script1_path = Path(__file__).parent.resolve() / "scripts/check_lamin_dev.py"
    subprocess.run(  # noqa: S602
        f"python {script1_path}",
        shell=True,
        check=True,
    )



import lamindb as ln

ln.examples.datasets.mini_immuno.define_features_labels()
adata = ln.examples.datasets.mini_immuno.get_dataset1(otype="AnnData")
schema = ln.examples.schemas.anndata_ensembl_gene_ids_and_valid_features_in_obs()
artifact = ln.Artifact.from_anndata(
    adata, key="examples/mini_immuno.h5ad", schema=schema
).save()
artifact.describe()



from __future__ import annotations

from lamindb_setup.core.upath import VALID_COMPOSITE_SUFFIXES, VALID_SIMPLE_SUFFIXES

# add new composite suffixes like so
VALID_COMPOSITE_SUFFIXES.update(
    {
        ".vitessce.json",
        ".ome.zarr",
    }
)
# can do the same for simple valid suffixes


class VALID_SUFFIXES:
    """Valid suffixes."""

    SIMPLE: set[str] = VALID_SIMPLE_SUFFIXES
    """Simple suffixes."""
    COMPOSITE: set[str] = VALID_COMPOSITE_SUFFIXES
    """Composite suffixes."""



from datetime import datetime

from django.db import models
from django.db.models import CASCADE, CharField, DateTimeField, ForeignKey, TextField

from .artifact import Artifact
from .collection import Collection
from .feature import Feature
from .project import Project
from .record import Record
from .run import Run, User
from .schema import Schema
from .sqlrecord import BaseSQLRecord, Branch, IsVersioned, Space
from .transform import Transform


class BlockMixin(IsVersioned):
    class Meta:
        abstract = True

    _len_full_uid: int = 20
    _len_stem_uid: int = 16

    id = models.BigAutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(
        editable=False, unique=True, db_index=True, max_length=_len_full_uid
    )
    """Universal id."""
    content: str = TextField()
    """Content of the block."""
    hash: str = CharField(max_length=22, db_index=True, null=True)
    """Content hash of the block."""
    type: str = CharField(
        max_length=22, db_index=True, default="mdpage", db_default="mdpage"
    )
    """Type of the block.

    Only current option is: "mdpage" (markdown page).
    """
    vertical_pos: float = models.FloatField(default=0, db_default=0, db_index=True)
    """The vertical position of the block in the GUI as a float.

    It can be used to act as a percentage to define absolute positioning or for ordering.
    """
    created_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Time of creation of record."""
    created_by: User = ForeignKey(
        "User", CASCADE, default=None, related_name="+", null=True
    )
    """Creator of block."""


class RootBlock(BlockMixin, BaseSQLRecord):
    """A root block for every registry that can appear at the top of the registry root block in the GUI."""

    class Meta:
        app_label = "lamindb"

    name: str = CharField(max_length=255, db_index=True)
    """The entity for which we want to create a block.

    Conventions are mostly to take the SQL table name:
        name = "instance" means instance
        name = "lamindb_artifact" means artifact
        name = "lamindb_transform" means transform
        name = "bionty_celltype" means bionty cell type
    """


class RecordBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to a record."""

    class Meta:
        app_label = "lamindb"

    record: Record = ForeignKey(Record, CASCADE, related_name="blocks")
    """The record to which the block is attached."""


class ArtifactBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to an artifact."""

    class Meta:
        app_label = "lamindb"

    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="blocks")
    """The artifact to which the block is attached."""


class TransformBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to a transform."""

    class Meta:
        app_label = "lamindb"

    transform: Transform = ForeignKey(
        Transform, CASCADE, related_name="blocks", null=True
    )
    """The transform to which the block is attached."""
    line_number: int | None = models.IntegerField(null=True)
    """The line number in the source code to which the block belongs."""


class RunBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to a run."""

    class Meta:
        app_label = "lamindb"

    run: Run = ForeignKey(Run, CASCADE, related_name="blocks")
    """The run to which the block is attached."""


class CollectionBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to a collection."""

    class Meta:
        app_label = "lamindb"

    collection: Collection = ForeignKey(
        Collection, CASCADE, related_name="blocks", null=True
    )
    """The collection to which the block is attached."""


class SchemaBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to a schema."""

    class Meta:
        app_label = "lamindb"

    schema: Schema = ForeignKey(Schema, CASCADE, related_name="blocks")
    """The schema to which the block is attached."""


class FeatureBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to a feature."""

    class Meta:
        app_label = "lamindb"

    feature: Feature = ForeignKey(Feature, CASCADE, related_name="blocks")
    """The feature to which the block is attached."""


class ProjectBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to a project."""

    class Meta:
        app_label = "lamindb"

    project: Project = ForeignKey(Project, CASCADE, related_name="blocks")
    """The project to which the block is attached."""


class SpaceBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to a space."""

    class Meta:
        app_label = "lamindb"

    space: Space = ForeignKey(Space, CASCADE, related_name="blocks")
    """The space to which the block is attached."""


class BranchBlock(BlockMixin, BaseSQLRecord):
    """An unstructured notes block that can be attached to a branch."""

    class Meta:
        app_label = "lamindb"

    branch: Branch = ForeignKey(Branch, CASCADE, related_name="blocks")
    """The branch to which the block is attached."""



# Generated by Django 5.2 on 2025-10-01 09:07

from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0127_alter_run_status_code_feature_dtype"),
    ]

    operations = [
        migrations.AddField(
            model_name="artifact",
            name="_real_key",
            field=lamindb.base.fields.CharField(
                blank=True, db_index=True, default=None, max_length=1024, null=True
            ),
        ),
    ]



"""Models library.

.. autosummary::
   :toctree: .

   BaseSQLRecord
   SQLRecord
   Registry
   BasicQuerySet
   QuerySet
   ArtifactSet
   LazyArtifact
   QueryManager
   SQLRecordList
   FeatureManager
   LabelManager
   IsVersioned
   CanCurate
   HasParents
   TracksRun
   TracksUpdates
   FeatureValue
   InspectResult
   ValidateFields
   SchemaOptionals

"""

# ruff: noqa: I001
from lamin_utils._inspect import InspectResult
from ._is_versioned import IsVersioned
from .can_curate import CanCurate
from .sqlrecord import (
    BaseSQLRecord,
    SQLRecord,
    Registry,
    Space,
    Branch,
    Migration,
    ValidateFields,
    format_field_value,
    record_repr,
    IsLink,
)
from .storage import Storage
from .transform import Transform
from .run import Run, TracksRun, TracksUpdates, current_run, User
from .feature import Feature, FeatureValue
from .schema import Schema
from .ulabel import ULabel

# should come last as it needs everything else
from .artifact import Artifact, LazyArtifact
from ._feature_manager import FeatureManager
from ._label_manager import LabelManager
from .collection import Collection, CollectionArtifact
from .project import Project, Reference
from .query_manager import QueryManager
from .query_set import BasicQuerySet, QuerySet, SQLRecordList
from .artifact_set import ArtifactSet
from .has_parents import HasParents
from datetime import datetime as _datetime

FeatureSet = Schema  # backward compat

# link models
from .artifact import ArtifactFeatureValue
from .project import (
    ArtifactProject,
    TransformProject,
    CollectionProject,
    ULabelProject,
    FeatureProject,
    SchemaProject,
    ArtifactReference,
    CollectionReference,
    RunProject,
    RecordProject,
    RecordReference,
    ReferenceRecord,
    ProjectRecord,
)
from .run import RunFeatureValue
from .schema import (
    SchemaFeature,
    ArtifactSchema,
    SchemaComponent,
    SchemaOptionals,
)
from .ulabel import ArtifactULabel, TransformULabel, RunULabel, CollectionULabel

from .record import (
    Record,
    RecordJson,
    RecordRecord,
    RecordULabel,
    RecordRun,
    RecordUser,
    RecordArtifact,
    ArtifactRecord,
)
from .block import (
    RootBlock,
    ArtifactBlock,
    TransformBlock,
    RecordBlock,
    CollectionBlock,
    RunBlock,
    SchemaBlock,
    ProjectBlock,
    BranchBlock,
    SpaceBlock,
)

LinkORM = IsLink  # backward compat
ParamValue = FeatureValue  # backward compat
ArtifactParamValue = ArtifactFeatureValue  # backward compat
RunParamValue = RunFeatureValue  # backward compat
Param = Feature  # backward compat
BasicRecord = BaseSQLRecord  # backward compat
